# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import logging
import asyncio
import pyuavcan.transport
from ._base import RedundantSession, RedundantSessionStatistics


_logger = logging.getLogger(__name__)


class RedundantFeedback(pyuavcan.transport.Feedback):
    """
    This is the output feedback extended with the reference to the inferior transport session
    that this feedback originates from.

    A redundant output session provides one feedback entry per inferior session;
    for example, if there are three inferiors in a redundant transport group,
    each outgoing transfer will generate three feedback entries
    (unless inferior sessions fail to provide their feedback entries for whatever reason).
    """

    def __init__(
        self, inferior_feedback: pyuavcan.transport.Feedback, inferior_session: pyuavcan.transport.OutputSession
    ):
        self._inferior_feedback = inferior_feedback
        self._inferior_session = inferior_session

    @property
    def original_transfer_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._inferior_feedback.original_transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._inferior_feedback.first_frame_transmission_timestamp

    @property
    def inferior_feedback(self) -> pyuavcan.transport.Feedback:
        """
        The original feedback instance from the inferior session.
        """
        assert isinstance(self._inferior_feedback, pyuavcan.transport.Feedback)
        return self._inferior_feedback

    @property
    def inferior_session(self) -> pyuavcan.transport.OutputSession:
        """
        The inferior session that generated this feedback entry.
        """
        assert isinstance(self._inferior_session, pyuavcan.transport.OutputSession)
        return self._inferior_session


class RedundantOutputSession(RedundantSession, pyuavcan.transport.OutputSession):
    """
    This is a composite of a group of :class:`pyuavcan.transport.OutputSession`.
    Every outgoing transfer is simply forked into each of the inferior sessions.
    The result aggregation policy is documented in :func:`send`.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly! Use the factory method instead.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._loop = loop
        self._finalizer: typing.Optional[typing.Callable[[], None]] = finalizer
        assert isinstance(self._specifier, pyuavcan.transport.OutputSessionSpecifier)
        assert isinstance(self._payload_metadata, pyuavcan.transport.PayloadMetadata)
        assert isinstance(self._loop, asyncio.AbstractEventLoop)
        assert callable(self._finalizer)

        self._inferiors: typing.List[pyuavcan.transport.OutputSession] = []
        self._feedback_handler: typing.Optional[typing.Callable[[RedundantFeedback], None]] = None
        self._idle_send_future: typing.Optional[asyncio.Future[None]] = None
        self._lock = asyncio.Lock()

        self._stat_transfers = 0
        self._stat_payload_bytes = 0
        self._stat_errors = 0
        self._stat_drops = 0

    def _add_inferior(self, session: pyuavcan.transport.Session) -> None:
        assert isinstance(session, pyuavcan.transport.OutputSession)
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        assert session.specifier == self.specifier and session.payload_metadata == self.payload_metadata
        if session not in self._inferiors:
            # Synchronize the feedback state.
            if self._feedback_handler is not None:
                self._enable_feedback_on_inferior(session)
            else:
                session.disable_feedback()
            # If and only if all went well, add the new inferior to the set.
            self._inferiors.append(session)
            # Unlock the pending transmission because now we have an inferior to work with.
            if self._idle_send_future is not None:
                self._idle_send_future.set_result(None)

    def _close_inferior(self, session_index: int) -> None:
        assert session_index >= 0, "Negative indexes may lead to unexpected side effects"
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        try:
            session = self._inferiors.pop(session_index)
        except LookupError:
            pass
        else:
            session.close()  # May raise.

    @property
    def inferiors(self) -> typing.Sequence[pyuavcan.transport.OutputSession]:
        return self._inferiors[:]

    def enable_feedback(self, handler: typing.Callable[[RedundantFeedback], None]) -> None:
        """
        The operation is atomic on all inferiors.
        If at least one inferior fails to enable feedback, all inferiors are rolled back into the disabled state.
        """
        self.disable_feedback()  # For state determinism.
        try:
            self._feedback_handler = handler
            for ses in self._inferiors:
                self._enable_feedback_on_inferior(ses)
        except Exception as ex:
            _logger.info("%s could not enable feedback, rolling back into the disabled state: %r", self, ex)
            self.disable_feedback()
            raise

    def disable_feedback(self) -> None:
        """
        The method implements the best-effort policy if any of the inferior sessions fail to disable feedback.
        """
        self._feedback_handler = None
        for ses in self._inferiors:
            try:
                ses.disable_feedback()
            except Exception as ex:
                _logger.exception("%s could not disable feedback on %r: %s", self, ses, ex)

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        """
        Sends the transfer via all of the inferior sessions concurrently.
        Returns when all of the inferior calls return and/or raise exceptions.
        Edge cases:

        - If there are no inferiors, the method will await until either the deadline is expired
          or an inferior(s) is (are) added. In the former case, the method returns False.
          In the latter case, the transfer is transmitted via the new inferior(s) using the remaining time
          until the deadline.

        - If at least one inferior succeeds, True is returned (logical OR).
          If the other inferiors raise exceptions, they are logged as errors and suppressed.

        - If all inferiors raise exceptions, the exception from the first one is propagated,
          the rest are logged as errors and suppressed.

        - If all inferiors time out, False is returned (logical OR).

        In other words, the error handling strategy is optimistic: if one inferior reported success,
        the call is assumed to have succeeded; best result is always returned.
        """
        if self._finalizer is None:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

        async with self._lock:  # Serialize access to the inferiors and the idle future.
            # It is required to create a local copy to prevent disruption of the logic when
            # the set of inferiors is changed in the background. Oh, Rust, where art thou.
            inferiors = list(self._inferiors)

            # This part is a bit tricky. If there are no inferiors, we have nowhere to send the transfer.
            # Instead of returning immediately, we hang out here until the deadline is expired hoping that
            # an inferior is added while we're waiting here.
            assert not self._idle_send_future
            if not inferiors and monotonic_deadline > self._loop.time():
                try:
                    _logger.debug("%s has no inferiors; suspending the send method...", self)
                    self._idle_send_future = self._loop.create_future()
                    try:
                        await asyncio.wait_for(self._idle_send_future, timeout=monotonic_deadline - self._loop.time())
                    except asyncio.TimeoutError:
                        pass
                    else:
                        self._idle_send_future.result()  # Collect the empty result to prevent asyncio from complaining.
                    # The set of inferiors may have been updated.
                    inferiors = list(self._inferiors)
                    _logger.debug(
                        "%s send method unsuspended; available inferiors: %r; remaining time: %f",
                        self,
                        inferiors,
                        monotonic_deadline - self._loop.time(),
                    )
                finally:
                    self._idle_send_future = None
            assert not self._idle_send_future

            if not inferiors:
                self._stat_drops += 1
                return False  # Still nothing.

            results = await asyncio.gather(
                *[ses.send(transfer, monotonic_deadline) for ses in inferiors], loop=self._loop, return_exceptions=True
            )
            assert results and len(results) == len(inferiors)
            _logger.debug("%s send results: %s", self, results)

            exceptions = [ex for ex in results if isinstance(ex, Exception)]

            # Result consolidation logic as described in the doc.
            if exceptions:
                # Taking great efforts to make the error message very understandable to the user.
                _logger.error(  # pylint: disable=logging-not-lazy
                    f"{self}: {len(exceptions)} of {len(results)} inferiors have failed: "
                    + ", ".join(f"{i}:{self._describe_send_result(r)}" for i, r in enumerate(results))
                )
                if len(exceptions) >= len(results):
                    self._stat_errors += 1
                    raise exceptions[0]

            if any(x is True for x in results):
                self._stat_transfers += 1
                self._stat_payload_bytes += sum(map(len, transfer.fragmented_payload))
                return True
            self._stat_drops += 1
            return False

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> RedundantSessionStatistics:
        """
        - ``transfers``     - the number of redundant transfers where at least ONE inferior succeeded (success count).
        - ``errors``        - the number of redundant transfers where ALL inferiors raised exceptions (failure count).
        - ``payload_bytes`` - the number of payload bytes in successful redundant transfers counted in ``transfers``.
        - ``drops``         - the number of redundant transfers where ALL inferiors timed out (timeout count).
        - ``frames``        - the total number of frames summed from all inferiors (i.e., replicated frame count).
          This value is invalidated when the set of inferiors is changed. The semantics may change later.
        """
        inferiors = [s.sample_statistics() for s in self._inferiors]
        return RedundantSessionStatistics(
            transfers=self._stat_transfers,
            frames=sum(s.frames for s in inferiors),
            payload_bytes=self._stat_payload_bytes,
            errors=self._stat_errors,
            drops=self._stat_drops,
            inferiors=inferiors,
        )

    def close(self) -> None:
        for s in self._inferiors:
            try:
                s.close()
            except Exception as ex:
                _logger.exception("%s could not close inferior %s: %s", self, s, ex)
        self._inferiors.clear()

        fin, self._finalizer = self._finalizer, None
        if fin is not None:
            fin()

    def _enable_feedback_on_inferior(self, inferior_session: pyuavcan.transport.OutputSession) -> None:
        def proxy(fb: pyuavcan.transport.Feedback) -> None:
            """
            Intercepts a feedback report from an inferior session,
            constructs a higher-level redundant feedback instance from it,
            and then passes it along to the higher-level handler.
            """
            if inferior_session not in self._inferiors:
                _logger.warning(
                    "%s got unexpected feedback %s from %s which is not a registered inferior. "
                    "The transport or its underlying software or hardware are probably misbehaving, "
                    "or this inferior has just been removed.",
                    self,
                    fb,
                    inferior_session,
                )
                return

            handler = self._feedback_handler
            if handler is not None:
                new_fb = RedundantFeedback(fb, inferior_session)
                try:
                    handler(new_fb)
                except Exception as ex:
                    _logger.exception("%s: Unhandled exception in the feedback handler %s: %s", self, handler, ex)
            else:
                _logger.debug("%s ignoring unattended feedback %r from %r", self, fb, inferior_session)

        inferior_session.enable_feedback(proxy)

    @staticmethod
    def _describe_send_result(result: typing.Union[bool, Exception]) -> str:
        if isinstance(result, Exception):
            return repr(result)
        if isinstance(result, bool):
            return "success" if result else "timeout"
        assert False


def _unittest_redundant_output() -> None:
    import time
    import pytest
    from pyuavcan.transport import Transfer, Timestamp, Priority, SessionStatistics, ResourceClosedError
    from pyuavcan.transport import TransferFrom
    from pyuavcan.transport.loopback import LoopbackTransport, LoopbackFeedback

    loop = asyncio.get_event_loop()
    await_ = loop.run_until_complete

    spec = pyuavcan.transport.OutputSessionSpecifier(pyuavcan.transport.MessageDataSpecifier(4321), None)
    spec_rx = pyuavcan.transport.InputSessionSpecifier(spec.data_specifier, None)
    meta = pyuavcan.transport.PayloadMetadata(30 * 1024 * 1024)

    ts = Timestamp.now()

    is_retired = False

    def retire() -> None:
        nonlocal is_retired
        is_retired = True

    ses = RedundantOutputSession(spec, meta, loop=loop, finalizer=retire)
    assert not is_retired
    assert ses.specifier is spec
    assert ses.payload_metadata is meta
    assert not ses.inferiors
    assert ses.sample_statistics() == RedundantSessionStatistics()

    # Transmit with an empty set of inferiors.
    time_before = loop.time()
    assert not await_(
        ses.send(
            Transfer(
                timestamp=ts,
                priority=Priority.IMMEDIATE,
                transfer_id=1234567890,
                fragmented_payload=[memoryview(b"abc")],
            ),
            loop.time() + 2.0,
        )
    )
    assert 1.0 < loop.time() - time_before < 5.0, "The method should have returned in about two seconds."
    assert ses.sample_statistics() == RedundantSessionStatistics(
        drops=1,
    )

    # Create inferiors.
    tr_a = LoopbackTransport(111)
    tr_b = LoopbackTransport(111)
    inf_a = tr_a.get_output_session(spec, meta)
    inf_b = tr_b.get_output_session(spec, meta)
    rx_a = tr_a.get_input_session(spec_rx, meta)
    rx_b = tr_b.get_input_session(spec_rx, meta)

    # Begin transmission, then add an inferior while it is in progress.
    async def add_inferior(inferior: pyuavcan.transport.OutputSession) -> None:
        _logger.debug("Test: sleeping before adding the inferior...")
        await asyncio.sleep(2.0)
        _logger.debug("Test: adding the inferior...")
        ses._add_inferior(inferior)  # pylint: disable=protected-access
        _logger.debug("Test: inferior has been added.")

    assert await_(
        asyncio.gather(
            # Start transmission here. It would stall for up to five seconds because no inferiors.
            ses.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.IMMEDIATE,
                    transfer_id=9876543210,
                    fragmented_payload=[memoryview(b"def")],
                ),
                loop.time() + 5.0,
            ),
            # While the transmission is stalled, add one inferior with a 2-sec delay. It will unlock the stalled task.
            add_inferior(inf_a),
            # Then make sure that the transmission has actually taken place about after two seconds from the start.
        )
    ), "Transmission should have succeeded"
    assert 1.0 < loop.time() - time_before < 5.0, "The method should have returned in about two seconds."
    assert ses.sample_statistics() == RedundantSessionStatistics(
        transfers=1,
        frames=1,
        payload_bytes=3,
        drops=1,
        inferiors=[
            SessionStatistics(
                transfers=1,
                frames=1,
                payload_bytes=3,
            ),
        ],
    )
    tf_rx = await_(rx_a.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 9876543210
    assert tf_rx.fragmented_payload == [memoryview(b"def")]
    assert None is await_(rx_b.receive(loop.time() + 0.1))

    # Enable feedback.
    feedback: typing.List[RedundantFeedback] = []
    ses.enable_feedback(feedback.append)
    assert await_(
        ses.send(
            Transfer(
                timestamp=ts,
                priority=Priority.LOW,
                transfer_id=555555555555,
                fragmented_payload=[memoryview(b"qwerty")],
            ),
            loop.time() + 1.0,
        )
    )
    assert ses.sample_statistics() == RedundantSessionStatistics(
        transfers=2,
        frames=2,
        payload_bytes=9,
        drops=1,
        inferiors=[
            SessionStatistics(
                transfers=2,
                frames=2,
                payload_bytes=9,
            ),
        ],
    )
    assert len(feedback) == 1
    assert feedback[0].inferior_session is inf_a
    assert feedback[0].original_transfer_timestamp == ts
    assert ts.system <= feedback[0].first_frame_transmission_timestamp.system <= time.time()
    assert ts.monotonic <= feedback[0].first_frame_transmission_timestamp.monotonic <= time.monotonic()
    assert isinstance(feedback[0].inferior_feedback, LoopbackFeedback)
    feedback.pop()
    assert not feedback
    tf_rx = await_(rx_a.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 555555555555
    assert tf_rx.fragmented_payload == [memoryview(b"qwerty")]
    assert None is await_(rx_b.receive(loop.time() + 0.1))

    # Add a new inferior and ensure that its feedback is auto-enabled!
    ses._add_inferior(inf_b)  # pylint: disable=protected-access
    assert ses.inferiors == [
        inf_a,
        inf_b,
    ]
    # Double-add has no effect.
    ses._add_inferior(inf_b)  # pylint: disable=protected-access
    assert ses.inferiors == [
        inf_a,
        inf_b,
    ]
    assert await_(
        ses.send(
            Transfer(
                timestamp=ts,
                priority=Priority.FAST,
                transfer_id=777777777777,
                fragmented_payload=[memoryview(b"fgsfds")],
            ),
            loop.time() + 1.0,
        )
    )
    assert ses.sample_statistics() == RedundantSessionStatistics(
        transfers=3,
        frames=3 + 1,
        payload_bytes=15,
        drops=1,
        inferiors=[
            SessionStatistics(
                transfers=3,
                frames=3,
                payload_bytes=15,
            ),
            SessionStatistics(
                transfers=1,
                frames=1,
                payload_bytes=6,
            ),
        ],
    )
    assert len(feedback) == 2
    assert feedback[0].inferior_session is inf_a
    assert feedback[0].original_transfer_timestamp == ts
    assert ts.system <= feedback[0].first_frame_transmission_timestamp.system <= time.time()
    assert ts.monotonic <= feedback[0].first_frame_transmission_timestamp.monotonic <= time.monotonic()
    assert isinstance(feedback[0].inferior_feedback, LoopbackFeedback)
    feedback.pop(0)
    assert len(feedback) == 1
    assert feedback[0].inferior_session is inf_b
    assert feedback[0].original_transfer_timestamp == ts
    assert ts.system <= feedback[0].first_frame_transmission_timestamp.system <= time.time()
    assert ts.monotonic <= feedback[0].first_frame_transmission_timestamp.monotonic <= time.monotonic()
    assert isinstance(feedback[0].inferior_feedback, LoopbackFeedback)
    feedback.pop()
    assert not feedback
    tf_rx = await_(rx_a.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 777777777777
    assert tf_rx.fragmented_payload == [memoryview(b"fgsfds")]
    tf_rx = await_(rx_b.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 777777777777
    assert tf_rx.fragmented_payload == [memoryview(b"fgsfds")]

    # Remove the first inferior.
    ses._close_inferior(0)  # pylint: disable=protected-access
    assert ses.inferiors == [inf_b]
    ses._close_inferior(1)  # Out of range, no effect.  # pylint: disable=protected-access
    assert ses.inferiors == [inf_b]
    # Make sure the removed inferior has been closed.
    assert not tr_a.output_sessions

    # Transmission test with the last inferior.
    assert await_(
        ses.send(
            Transfer(
                timestamp=ts,
                priority=Priority.HIGH,
                transfer_id=88888888888888,
                fragmented_payload=[memoryview(b"hedgehog")],
            ),
            loop.time() + 1.0,
        )
    )
    assert ses.sample_statistics().transfers == 4
    # We don't check frames because this stat metric is computed quite clumsily atm, this may change later.
    assert ses.sample_statistics().payload_bytes == 23
    assert ses.sample_statistics().drops == 1
    assert ses.sample_statistics().inferiors == [
        SessionStatistics(
            transfers=2,
            frames=2,
            payload_bytes=14,
        ),
    ]
    assert len(feedback) == 1
    assert feedback[0].inferior_session is inf_b
    assert feedback[0].original_transfer_timestamp == ts
    assert ts.system <= feedback[0].first_frame_transmission_timestamp.system <= time.time()
    assert ts.monotonic <= feedback[0].first_frame_transmission_timestamp.monotonic <= time.monotonic()
    assert isinstance(feedback[0].inferior_feedback, LoopbackFeedback)
    feedback.pop()
    assert not feedback
    assert None is await_(rx_a.receive(loop.time() + 1))
    tf_rx = await_(rx_b.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 88888888888888
    assert tf_rx.fragmented_payload == [memoryview(b"hedgehog")]

    # Disable the feedback.
    ses.disable_feedback()
    # A diversion - enable the feedback in the inferior and make sure it's not propagated.
    ses._enable_feedback_on_inferior(inf_b)  # pylint: disable=protected-access
    assert await_(
        ses.send(
            Transfer(
                timestamp=ts,
                priority=Priority.OPTIONAL,
                transfer_id=666666666666666,
                fragmented_payload=[memoryview(b"horse")],
            ),
            loop.time() + 1.0,
        )
    )
    assert ses.sample_statistics().transfers == 5
    # We don't check frames because this stat metric is computed quite clumsily atm, this may change later.
    assert ses.sample_statistics().payload_bytes == 28
    assert ses.sample_statistics().drops == 1
    assert ses.sample_statistics().inferiors == [
        SessionStatistics(
            transfers=3,
            frames=3,
            payload_bytes=19,
        ),
    ]
    assert not feedback
    assert None is await_(rx_a.receive(loop.time() + 1))
    tf_rx = await_(rx_b.receive(loop.time() + 1))
    assert isinstance(tf_rx, TransferFrom)
    assert tf_rx.transfer_id == 666666666666666
    assert tf_rx.fragmented_payload == [memoryview(b"horse")]

    # Retirement.
    assert not is_retired
    ses.close()
    assert is_retired
    # Make sure the inferiors have been closed.
    assert not tr_a.output_sessions
    assert not tr_b.output_sessions
    # Idempotency.
    is_retired = False
    ses.close()
    assert not is_retired

    # Use after close.
    with pytest.raises(ResourceClosedError):
        await_(
            ses.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.OPTIONAL,
                    transfer_id=1111111111111,
                    fragmented_payload=[memoryview(b"cat")],
                ),
                loop.time() + 1.0,
            )
        )

    assert None is await_(rx_a.receive(loop.time() + 1))
    assert None is await_(rx_b.receive(loop.time() + 1))


def _unittest_redundant_output_exceptions(caplog: typing.Any) -> None:
    import pytest
    from pyuavcan.transport import Transfer, Timestamp, Priority, SessionStatistics
    from pyuavcan.transport import TransferFrom
    from pyuavcan.transport.loopback import LoopbackTransport

    loop = asyncio.get_event_loop()
    await_ = loop.run_until_complete

    spec = pyuavcan.transport.OutputSessionSpecifier(pyuavcan.transport.MessageDataSpecifier(4321), None)
    spec_rx = pyuavcan.transport.InputSessionSpecifier(spec.data_specifier, None)
    meta = pyuavcan.transport.PayloadMetadata(30 * 1024 * 1024)

    ts = Timestamp.now()

    is_retired = False

    def retire() -> None:
        nonlocal is_retired
        is_retired = True

    ses = RedundantOutputSession(spec, meta, loop=loop, finalizer=retire)
    assert not is_retired
    assert ses.specifier is spec
    assert ses.payload_metadata is meta
    assert not ses.inferiors
    assert ses.sample_statistics() == RedundantSessionStatistics()

    tr_a = LoopbackTransport(111)
    tr_b = LoopbackTransport(111)
    inf_a = tr_a.get_output_session(spec, meta)
    inf_b = tr_b.get_output_session(spec, meta)
    rx_a = tr_a.get_input_session(spec_rx, meta)
    rx_b = tr_b.get_input_session(spec_rx, meta)
    ses._add_inferior(inf_a)  # pylint: disable=protected-access
    ses._add_inferior(inf_b)  # pylint: disable=protected-access

    # Transmission with exceptions.
    # If at least one transmission succeeds, the call succeeds.
    with caplog.at_level(logging.CRITICAL, logger=__name__):
        inf_a.exception = RuntimeError("INTENDED EXCEPTION")
        assert await_(
            ses.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.FAST,
                    transfer_id=444444444444,
                    fragmented_payload=[memoryview(b"INTENDED EXCEPTION")],
                ),
                loop.time() + 1.0,
            )
        )
        assert ses.sample_statistics() == RedundantSessionStatistics(
            transfers=1,
            frames=1,
            payload_bytes=len("INTENDED EXCEPTION"),
            errors=0,
            drops=0,
            inferiors=[
                SessionStatistics(
                    transfers=0,
                    frames=0,
                    payload_bytes=0,
                ),
                SessionStatistics(
                    transfers=1,
                    frames=1,
                    payload_bytes=len("INTENDED EXCEPTION"),
                ),
            ],
        )
        assert None is await_(rx_a.receive(loop.time() + 1))
        tf_rx = await_(rx_b.receive(loop.time() + 1))
        assert isinstance(tf_rx, TransferFrom)
        assert tf_rx.transfer_id == 444444444444
        assert tf_rx.fragmented_payload == [memoryview(b"INTENDED EXCEPTION")]

        # Transmission timeout.
        # One times out, one raises an exception --> the result is timeout.
        inf_b.should_timeout = True
        assert not await_(
            ses.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.FAST,
                    transfer_id=2222222222222,
                    fragmented_payload=[memoryview(b"INTENDED EXCEPTION")],
                ),
                loop.time() + 1.0,
            )
        )
        assert ses.sample_statistics().transfers == 1
        assert ses.sample_statistics().payload_bytes == len("INTENDED EXCEPTION")
        assert ses.sample_statistics().errors == 0
        assert ses.sample_statistics().drops == 1
        assert None is await_(rx_a.receive(loop.time() + 1))
        assert None is await_(rx_b.receive(loop.time() + 1))

        # Transmission with exceptions.
        # If all transmissions fail, the call fails.
        inf_b.exception = RuntimeError("INTENDED EXCEPTION")
        with pytest.raises(RuntimeError, match="INTENDED EXCEPTION"):
            assert await_(
                ses.send(
                    Transfer(
                        timestamp=ts,
                        priority=Priority.FAST,
                        transfer_id=3333333333333,
                        fragmented_payload=[memoryview(b"INTENDED EXCEPTION")],
                    ),
                    loop.time() + 1.0,
                )
            )
        assert ses.sample_statistics().transfers == 1
        assert ses.sample_statistics().payload_bytes == len("INTENDED EXCEPTION")
        assert ses.sample_statistics().errors == 1
        assert ses.sample_statistics().drops == 1
        assert None is await_(rx_a.receive(loop.time() + 1))
        assert None is await_(rx_b.receive(loop.time() + 1))

    # Retirement.
    assert not is_retired
    ses.close()
    assert is_retired
    # Make sure the inferiors have been closed.
    assert not tr_a.output_sessions
    assert not tr_b.output_sessions
    # Idempotency.
    is_retired = False
    ses.close()
    assert not is_retired



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import struct
import dataclasses
import pyuavcan


@dataclasses.dataclass(frozen=True, repr=False)
class UDPFrame(pyuavcan.transport.commons.high_overhead_transport.Frame):
    """
    The header format is up to debate until it's frozen in Specification.

    An important thing to keep in mind is that the minimum size of an UDP/IPv4 payload when transferred over
    100M Ethernet is 18 bytes, due to the minimum Ethernet frame size limit. That is, if the application
    payload requires less space, the missing bytes will be padded out to the minimum size.

    The current header format enables encoding by trivial memory aliasing on any conventional little-endian platform::

        struct Header {
            uint8_t  version;
            uint8_t  priority;
            uint16_t _reserved_a;   // Set to zero when encoding, ignore when decoding.
            uint32_t frame_index_eot;
            uint64_t transfer_id;
            uint64_t _reserved_b;   // Set to zero when encoding, ignore when decoding.
        };
        static_assert(sizeof(struct Header) == 24, "Invalid layout");

    If you have any feedback concerning the frame format, please bring it to
    https://forum.uavcan.org/t/alternative-transport-protocols/324.

    +---------------+---------------+---------------+-----------------+------------------+
    |**MAC header** | **IP header** |**UDP header** |**UAVCAN header**|**UAVCAN payload**|
    +---------------+---------------+---------------+-----------------+------------------+
    |                                               |    Layers modeled by this type     |
    +-----------------------------------------------+------------------------------------+
    """

    _HEADER_FORMAT = struct.Struct("<BB2xIQ8x")
    _VERSION = 0

    TRANSFER_ID_MASK = 2 ** 64 - 1
    INDEX_MASK = 2 ** 31 - 1

    def __post_init__(self) -> None:
        if not isinstance(self.priority, pyuavcan.transport.Priority):
            raise TypeError(f"Invalid priority: {self.priority}")  # pragma: no cover

        if not (0 <= self.transfer_id <= self.TRANSFER_ID_MASK):
            raise ValueError(f"Invalid transfer-ID: {self.transfer_id}")

        if not (0 <= self.index <= self.INDEX_MASK):
            raise ValueError(f"Invalid frame index: {self.index}")

        if not isinstance(self.payload, memoryview):
            raise TypeError(f"Bad payload type: {type(self.payload).__name__}")  # pragma: no cover

    def compile_header_and_payload(self) -> typing.Tuple[memoryview, memoryview]:
        """
        Compiles the UDP frame header and returns it as a read-only memoryview along with the payload, separately.
        The caller is supposed to handle the header and the payload independently.
        The reason is to avoid unnecessary data copying in the user space,
        allowing the caller to rely on the vectorized IO API instead (sendmsg).
        """
        header = self._HEADER_FORMAT.pack(
            self._VERSION, int(self.priority), self.index | ((1 << 31) if self.end_of_transfer else 0), self.transfer_id
        )
        return memoryview(header), self.payload

    @staticmethod
    def parse(image: memoryview) -> typing.Optional[UDPFrame]:
        try:
            version, int_priority, frame_index_eot, transfer_id = UDPFrame._HEADER_FORMAT.unpack_from(image)
        except struct.error:
            return None
        if version == UDPFrame._VERSION:
            # noinspection PyArgumentList
            return UDPFrame(
                priority=pyuavcan.transport.Priority(int_priority),
                transfer_id=transfer_id,
                index=(frame_index_eot & UDPFrame.INDEX_MASK),
                end_of_transfer=bool(frame_index_eot & (UDPFrame.INDEX_MASK + 1)),
                payload=image[UDPFrame._HEADER_FORMAT.size :],
            )
        return None


# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------


def _unittest_udp_frame_compile() -> None:
    from pyuavcan.transport import Priority
    from pytest import raises

    _ = UDPFrame(priority=Priority.LOW, transfer_id=0, index=0, end_of_transfer=False, payload=memoryview(b""))

    with raises(ValueError):
        _ = UDPFrame(
            priority=Priority.LOW, transfer_id=2 ** 64, index=0, end_of_transfer=False, payload=memoryview(b"")
        )

    with raises(ValueError):
        _ = UDPFrame(
            priority=Priority.LOW, transfer_id=0, index=2 ** 31, end_of_transfer=False, payload=memoryview(b"")
        )

    # Multi-frame, not the end of the transfer.
    assert (
        memoryview(
            b"\x00\x06\x00\x00"
            b"\r\xf0\xdd\x00"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
        ),
        memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame(
        priority=Priority.SLOW,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0x_0DD_F00D,
        end_of_transfer=False,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ).compile_header_and_payload()

    # Multi-frame, end of the transfer.
    assert (
        memoryview(
            b"\x00\x07\x00\x00"
            b"\r\xf0\xdd\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
        ),
        memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame(
        priority=Priority.OPTIONAL,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0x_0DD_F00D,
        end_of_transfer=True,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ).compile_header_and_payload()

    # Single-frame.
    assert (
        memoryview(
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
        ),
        memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame(
        priority=Priority.EXCEPTIONAL,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0,
        end_of_transfer=True,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ).compile_header_and_payload()


def _unittest_udp_frame_parse() -> None:
    from pyuavcan.transport import Priority

    for size in range(16):
        assert None is UDPFrame.parse(memoryview(bytes(range(size))))

    # Multi-frame, not the end of the transfer.
    assert UDPFrame(
        priority=Priority.SLOW,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0x_0DD_F00D,
        end_of_transfer=False,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame.parse(
        memoryview(
            b"\x00\x06\x00\x00"
            b"\r\xf0\xdd\x00"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"Well, I got here the same way the coin did."
        ),
    )

    # Multi-frame, end of the transfer.
    assert UDPFrame(
        priority=Priority.OPTIONAL,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0x_0DD_F00D,
        end_of_transfer=True,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame.parse(
        memoryview(
            b"\x00\x07\x00\x00"
            b"\r\xf0\xdd\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"Well, I got here the same way the coin did."
        ),
    )

    # Single-frame.
    assert UDPFrame(
        priority=Priority.EXCEPTIONAL,
        transfer_id=0x_DEAD_BEEF_C0FFEE,
        index=0,
        end_of_transfer=True,
        payload=memoryview(b"Well, I got here the same way the coin did."),
    ) == UDPFrame.parse(
        memoryview(
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"Well, I got here the same way the coin did."
        ),
    )

    # Too short.
    assert None is UDPFrame.parse(
        memoryview(
            b"\x00\x07\x00\x00"
            b"\r\xf0\xdd\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
        )[:-1],
    )
    # Bad version.
    assert None is UDPFrame.parse(
        memoryview(
            b"\x01\x07\x00\x00"
            b"\r\xf0\xdd\x80"
            b"\xee\xff\xc0\xef\xbe\xad\xde\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
        ),
    )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
from ._serializer import Serializer as Serializer
from ._deserializer import Deserializer as Deserializer

if sys.byteorder != "little":  # pragma: no cover
    raise RuntimeError(
        "BIG-ENDIAN PLATFORMS ARE NOT YET SUPPORTED. "
        "The current serialization code assumes that the native byte order is little-endian. Since UAVCAN uses "
        "little-endian byte order in its serialized data representations, this assumption allows us to bypass data "
        "transformation in many cases, resulting in zero-cost serialization and deserialization. "
        "Big-endian platforms are unable to take advantage of that, requiring byte swapping for multi-byte entities; "
        "fortunately, nowadays such platforms are uncommon. If you need to use this library on a big-endian platform, "
        "please implement the missing code and submit a pull request to the upstream, then remove this exception."
    )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Redundant pseudo-transport overview
+++++++++++++++++++++++++++++++++++

Native support for redundant transports is one of the core features of UAVCAN.
The class :class:`RedundantTransport` implements this feature within PyUAVCAN.
It works by aggregating zero or more instances of :class:`pyuavcan.transport.Transport`
into a *composite* that implements the redundant transport management logic as defined in the UAVCAN specification:

- Every outgoing transfer is replicated into all of the available redundant interfaces.
- Incoming transfers are deduplicated so that the local node receives at most one copy of each unique transfer
  received from the bus.

There exist two approaches to implementing transport-layer redundancy.
The differences are confined to the specifics of a particular implementation, they are not manifested on the bus
-- nodes exhibit identical behavior regardless of the chosen strategy:

- **Frame-level redundancy.**
  In this case, multiple redundant interfaces are managed by the same transport state machine.
  This strategy is more efficient in the sense of computing power and memory resources required to
  accommodate a given amount of networking workload compared to the alternative.
  Its limitation is that the redundant transports shall implement the same protocol (e.g., CAN),
  and all involved transports shall be configured to use the same MTU.

- **Transfer-level redundancy.**
  In this case, redundant interfaces are managed one level of abstraction higher:
  not at the level of separate *transport frames*, but at the level of complete *UAVCAN transfers*
  (if these terms sound unfamiliar, please read the UAVCAN specification).
  This approach complicates the data flow inside the library, but it supports *dissimilar transport redundancy*,
  allowing one to aggregate transports implementing different protocols (e.g., UDP with serial,
  possibly with different MTU).
  Dissimilar redundancy is often sought in high-reliability/safety-critical applications,
  as reviewed in https://forum.uavcan.org/t/557.

In accordance with its design goals, PyUAVCAN implements the transfer-level redundancy management strategy
since it offers greater flexibility and a wider set of available design options.
It is expected though that real-time embedded applications may often find frame-level redundancy preferable.

This implementation uses the term *inferior* to refer to a member of a redundant group:

- *Inferior transport* is a transport that belongs to a redundant transport group.
- *Inferior session* is a transport session that is owned by an inferior transport.

Whenever a redundant transport is requested to construct a new session,
it does so by initializing an instance of :class:`RedundantInputSession` or :class:`RedundantOutputSession`.
The constructed instance then holds a set of inferior sessions, one from each inferior transport,
all sharing the same session specifier (:class:`pyuavcan.transport.SessionSpecifier`).
The resulting relationship between inferior transports and inferior sessions can be conceptualized
as a matrix where columns represent inferior transports and rows represent sessions:

+-----------+---------------+---------------+---------------+---------------+
|           |  Transport 0  |  Transport 1  |      ...      |  Transport M  |
+===========+===============+===============+===============+===============+
| Session 0 |     S0T0      |     S0T1      |      ...      |     S0Tm      |
+-----------+---------------+---------------+---------------+---------------+
| Session 1 |     S1T0      |     S1T1      |      ...      |     S1Tm      |
+-----------+---------------+---------------+---------------+---------------+
|    ...    |      ...      |      ...      |      ...      |      ...      |
+-----------+---------------+---------------+---------------+---------------+
| Session N |     SnT0      |     SnT1      |      ...      |     SnTm      |
+-----------+---------------+---------------+---------------+---------------+

Attachment/detachment of a transport is modeled as an addition/removal of a column;
likewise, construction/retirement of a session is modeled as an addition/removal of a row.
While the construction of a row or a column is in progress, the matrix resides in an inconsistent state.
If any error occurs in the process, the matrix is rolled back to the previous consistent state,
and the already-constructed sessions of the new vector are retired.

Existing redundant sessions retain validity across any changes in the matrix configuration.
Logic that relies on a redundant instance is completely shielded from any changes in the underlying transport
configuration, meaning that the entire underlying transport structure may be swapped out with a completely
different one without affecting the higher levels.
A practical extreme case is where a redundant transport is constructed with zero inferior transports,
its session instances are configured, and the inferior transports are added later.
This is expected to be useful for long-running applications that have to retain the presentation-level structure
across changes in the transport configuration done on-the-fly without stopping the application.

Since the redundant transport itself also implements the interface :class:`pyuavcan.transport.Transport`,
it technically could be used as an inferior of another redundant transport instance,
although the practicality of such arrangement is questionable.
Attaching a redundant transport as an inferior of itself is expressly prohibited and results in an error.


Inferior aggregation restrictions
+++++++++++++++++++++++++++++++++

Transports are categorized into one of the following two categories by the value of their transfer-ID (TID) modulo
(i.e., the transfer-ID overflow period).

Transports where the set of transfer-ID values contains less than 2**48 (``0x_1_0000_0000_0000``)
distinct elements are said to have *cyclic transfer-ID*.
In such transports, the value of the transfer-ID increases steadily starting from zero,
incremented once per emitted transfer, until the highest value is reached,
then the value is wrapped over to zero::

    modulo
         /|   /|   /|
        / |  / |  / |
       /  | /  | /  | /
      /   |/   |/   |/
    0 ----------------->
            time


Transports where the set of transfer-ID values is larger are said to have *monotonic transfer-ID*.
In such transports, the set is considered to be large enough to be inexhaustible for any practical application,
hence a wrap-over to zero is expected to never occur.
(For example, a UAVCAN/UDP transport operating over a 10 GbE link at the theoretical throughput limit of
14.9 million transfers per second will exhaust the set in approx. 153 years in the worst case.)

Monotonic transports impose a higher data overhead per frame due to the requirement to accommodate a
sufficiently wide integer field for the transfer-ID value.
Their advantage is that transfer-ID values carried over inferior transports of a redundant group are guaranteed
to remain in-phase for the entire lifetime of the network.
The importance of this guarantee can be demonstrated with the following counter-example of two transports
leveraging different transfer-ID modulo for the same session,
where the unambiguous mapping between their transfer-ID values is lost
with the beginning of the epoch B1 after the first overflow::

    A0    A1    A2    A3
        /|    /|    /|
       / |   / |   / |   /
      /  |  /  |  /  |  /
     /   | /   | /   | /
    /    |/    |/    |/

    B0   B1   B2   B3   B4
       /|   /|   /|   /|
      / |  / |  / |  / |
     /  | /  | /  | /  | /
    /   |/   |/   |/   |/
    ---------------------->
             time

The phase ambiguity of cyclic-TID transports results in the following hard requirements:

1. Inferior transports under the same redundant transport instance shall belong to the same TID monotonicity category:
   either all cyclic or all monotonic.
2. In the case where the inferiors utilize cyclic TID counters, the TID modulo shall be identical for all inferiors.

The implementation raises an error if an attempt is made to violate any of the above requirements.
The TID monotonicity category of an inferior is determined by querying
:attr:`pyuavcan.transport.Transport.protocol_parameters`.


Transmission
++++++++++++

As stated in the Specification, every emitted transfer shall be replicated into all available redundant interfaces.
The rest of the logic does not concern wire compatibility, and hence it is implementation-defined.

This implementation applies an optimistic result aggregation policy where it considers a transmission successful
if at least one inferior was able to successfully complete it.
The handling of time-outs, exceptions, and other edge cases is described in detail in the documentation for
:class:`RedundantOutputSession`.

Every outgoing transfer will be serialized and transmitted by each inferior independently from each other.
This may result in different number of transport frames emitted if the inferiors are configured to use
different MTU, or if they implement different transport protocols.

Inferiors compute the modulus of the transfer-ID according to the protocol they implement
independently from each other;
however, despite the independent computation, it is guaranteed that they will always arrive at the same
final transfer-ID value thanks to the aggregation restrictions introduced earlier.
This guarantee is paramount for service calls, because UAVCAN requires the caller to match a service response
with the appropriate request state by comparing its transfer-ID value,
which in turn requires that the logic that performs such matching is aware about the transfer-ID modulo in use.


Reception
+++++++++

Received transfers need to be deduplicated (dereplicated) so that the higher layers of the protocol stack
would not receive each unique transfer more than once (as demanded by the Specification).

Transfer reception and deduplication are managed by the class :class:`RedundantInputSession`.
There exist two deduplication strategies, chosen automatically depending on the TID monotonicity category
of the inferiors
(as described earlier, it is enforced that all inferiors in a redundant group belong to the same
TID monotonicity category).

The cyclic-TID deduplication strategy picks a transport interface at random and stays with it as long as
the interface keeps delivering transfers.
If the currently used interface ceases to deliver transfers, the strategy may switch to another one,
thus manifesting the automatic fail-over.
The cyclic-TID strategy cannot utilize more than one interface simultaneously due to the risk of
transfer duplication induced by a possible transport latency disbalance
(this is discussed at https://github.com/UAVCAN/specification/issues/8 and in the Specification).

The monotonic-TID deduplication strategy always picks the first transfer to arrive.
This approach provides instant fail-over in the case of an interface failure and
ensures that the worst case transfer latency is bounded by the latency of the best-performing transport.

The following two swim lane diagrams should illustrate the difference.
First, the case of cyclic-TID::

    A   B     Deduplicated
    |   |     |
    T0  |     T0     <-- First transfer received from transport A.
    T1  T0    T1     <-- Transport B is auto-assigned as a back-up.
    T2  T1    T2     <-- Up to this point the transport functions normally.
    X   T2    |      <-- Transport A fails here.
        T3    |      <-- Valid transfers from transport B are ignored due to the mandatory fail-over delay.
        ...   |
        Tn    Tn     <-- After the delay, the deduplicator switches over to the back-up transport.
        Tn+1  Tn+1   <-- Now, the roles of the back-up transport and the main transport are swapped.
        Tn+2  Tn+2

Monotonic-TID::

    A   B     Deduplicated
    |   |     |
    T0  |     T0    <-- The monotonic-TID strategy always picks the first transfer to arrive.
    T1  T0    T1    <-- All available interfaces are always considered.
    T2  T1    T2    <-- The result is that the transfer latency is defined by the best-performing transport.
    |   T2    |     <-- Here, the latency of transport A has increased temporarily.
    |   T3    T3    <-- The deduplication strategy reacts by picking the next transfer from transport B.
    T3  X     |     <-- Shall one transport fail, the deduplication strategy fails over immediately.
    T4        T4

Anonymous transfers are a special case:
a deduplicator has to keep local state per session in order to perform its functions;
since anonymous transfers are fundamentally stateless, they are always accepted unconditionally.
The implication is that redundant transfers may be replicated.
This behavior is due to the design of the protocol and is not specific to this implementation.


Inheritance diagram
+++++++++++++++++++

.. inheritance-diagram:: pyuavcan.transport.redundant._redundant_transport
                         pyuavcan.transport.redundant._error
                         pyuavcan.transport.redundant._session._base
                         pyuavcan.transport.redundant._session._input
                         pyuavcan.transport.redundant._session._output
   :parts: 1


Usage
+++++

A freshly constructed redundant transport is empty.
Redundant transport instances are intentionally designed to be very mutable,
allowing one to reconfigure them freely on-the-fly to support the needs of highly dynamic applications.
Such flexibility allows one to do things that are illegal per the UAVCAN specification,
such as changing the node-ID while the node is running, so beware.

>>> tr = RedundantTransport()
>>> tr.inferiors  # By default, there are none.
[]

It is possible to begin creating session instances immediately, before configuring the inferiors.
Any future changes will update all dependent session instances automatically.

>>> from pyuavcan.transport import OutputSessionSpecifier, InputSessionSpecifier, MessageDataSpecifier
>>> from pyuavcan.transport import PayloadMetadata, Transfer, Timestamp, Priority, ProtocolParameters
>>> pm = PayloadMetadata(1024)
>>> s0 = tr.get_output_session(OutputSessionSpecifier(MessageDataSpecifier(2345), None), pm)
>>> s0.inferiors    # No inferior transports; hence, no inferior sessions.
[]

If we attempted to transmit or receive a transfer while there are no inferiors, the call would just time out.

In this example, we will be experimenting with the loopback transport.
Below we are attaching a new inferior transport instance; the session instances are updated automatically.

>>> from pyuavcan.transport.loopback import LoopbackTransport
>>> lo_0 = LoopbackTransport(local_node_id=42)
>>> tr.attach_inferior(lo_0)
>>> tr.inferiors
[LoopbackTransport(...)]
>>> s0.inferiors
[LoopbackOutputSession(...)]

Add another inferior and another session:

>>> lo_1 = LoopbackTransport(local_node_id=42)
>>> tr.attach_inferior(lo_1)
>>> s1 = tr.get_input_session(InputSessionSpecifier(MessageDataSpecifier(2345), None), pm)
>>> len(tr.inferiors)
2
>>> len(s0.inferiors)  # Updated automatically.
2
>>> len(s1.inferiors)
2
>>> assert tr.inferiors[0].output_sessions[0] is s0.inferiors[0]    # Navigating the session matrix.
>>> assert tr.inferiors[1].output_sessions[0] is s0.inferiors[1]
>>> assert tr.inferiors[0].input_sessions[0] is s1.inferiors[0]
>>> assert tr.inferiors[1].input_sessions[0] is s1.inferiors[1]

A simple exchange test (remember this is a loopback, so we get back whatever we send):

>>> await_ = tr.loop.run_until_complete
>>> await_(s0.send(Transfer(Timestamp.now(), Priority.LOW, 1111, fragmented_payload=[]), tr.loop.time() + 1.0))
True
>>> await_(s1.receive(tr.loop.time() + 1.0))
RedundantTransferFrom(..., transfer_id=1111, fragmented_payload=[], ...)

Inject a failure into one inferior.
The redundant transport will continue to function with the other inferior; an error message will be logged:

.. The 'doctest: +SKIP' is needed because PyTest is broken. If a failure is actually injected,
.. the transport will be logging errors, which in turn break the PyTest's doctest plugin.
.. This is a known bug which is documented here: https://github.com/pytest-dev/pytest/issues/5908.
.. When that is fixed (I suppose it should be by PyTest v6?), please, remove this comment and the 'doctest: +SKIP'.

>>> lo_0.output_sessions[0].exception = RuntimeError('Injected failure')  # doctest: +SKIP
>>> await_(s0.send(Transfer(Timestamp.now(), Priority.LOW, 1112, fragmented_payload=[]), tr.loop.time() + 1.0))
True
>>> await_(s1.receive(tr.loop.time() + 1.0))   # Still works.
RedundantTransferFrom(..., transfer_id=1112, fragmented_payload=[], ...)

Inferiors that are no longer needed can be detached.
The redundant transport cleans up after itself by closing all inferior sessions in the detached transport.

>>> tr.detach_inferior(lo_0)
>>> len(tr.inferiors)   # Yup, removed.
1
>>> len(s0.inferiors)   # And the existing session instances are updated.
1
>>> len(s1.inferiors)   # Indeed they are.
1

One cannot mix inferiors with incompatible TID monotonicity or different node-ID.
For example, it is not possible to use CAN with UDP in the same redundant group.

>>> lo_0 = LoopbackTransport(local_node_id=42)
>>> lo_0.protocol_parameters = ProtocolParameters(transfer_id_modulo=32, max_nodes=128, mtu=8)
>>> tr.attach_inferior(lo_0)                        # TID monotonicity mismatch.    #doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
    ...
InconsistentInferiorConfigurationError: The new inferior shall use monotonic transfer-ID counters...
>>> tr.attach_inferior(LoopbackTransport(local_node_id=None))  # Node-ID mismatch.  #doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
    ...
InconsistentInferiorConfigurationError: The inferior has a different node-ID...

The parameters of a redundant transport are computed from the inferiors.
If the inferior set is changed, the transport parameters may also be changed.
This may create unexpected complications because parameters of real transports are generally immutable,
so it is best to avoid unnecessary runtime transformations unless required by the business logic.

>>> tr.local_node_id
42
>>> tr.protocol_parameters
ProtocolParameters(...)
>>> tr.close()                  # All inferiors and all sessions are closed.
>>> tr.inferiors
[]
>>> tr.local_node_id is None
True
>>> tr.protocol_parameters
ProtocolParameters(transfer_id_modulo=0, max_nodes=0, mtu=0)

A redundant transport can be used with just one inferior to implement ad-hoc PnP allocation as follows:
the transport is set up with an anonymous inferior which is disposed of upon completing the allocation procedure;
the new inferior is then installed in the place of the old one configured to use the newly allocated node-ID value.
"""

from ._redundant_transport import RedundantTransport as RedundantTransport
from ._redundant_transport import RedundantTransportStatistics as RedundantTransportStatistics

from ._session import RedundantSession as RedundantSession
from ._session import RedundantInputSession as RedundantInputSession
from ._session import RedundantOutputSession as RedundantOutputSession

from ._session import RedundantSessionStatistics as RedundantSessionStatistics
from ._session import RedundantFeedback as RedundantFeedback

from ._error import InconsistentInferiorConfigurationError as InconsistentInferiorConfigurationError

from ._tracer import RedundantCapture as RedundantCapture
from ._tracer import RedundantDuplicateTransferTrace as RedundantDuplicateTransferTrace
from ._tracer import RedundantTracer as RedundantTracer



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import dataclasses
import pyuavcan


@dataclasses.dataclass(frozen=True)
class Frame:
    """
    The base class of a high-overhead-transport frame.
    It is used with the common transport algorithms defined in this module.
    Concrete transport implementations should make their transport-specific frame dataclasses inherit from this class.
    Derived types are recommended to not override ``__repr__()``.
    """

    priority: pyuavcan.transport.Priority
    """
    Transfer priority should be the same for all frames within the transfer.
    """

    transfer_id: int
    """
    Transfer-ID is incremented whenever a transfer under a specific session-specifier is emitted.
    Always non-negative.
    """

    index: int
    """
    Index of the frame within its transfer, starting from zero. Always non-negative.
    """

    end_of_transfer: bool
    """
    True for the last frame within the transfer.
    """

    payload: memoryview
    """
    The data carried by the frame. Multi-frame transfer payload is suffixed with its CRC32C. May be empty.
    """

    def __post_init__(self) -> None:
        if not isinstance(self.priority, pyuavcan.transport.Priority):
            raise TypeError(f"Invalid priority: {self.priority}")

        if self.transfer_id < 0:
            raise ValueError(f"Invalid transfer-ID: {self.transfer_id}")

        if self.index < 0:
            raise ValueError(f"Invalid frame index: {self.index}")

        if not isinstance(self.end_of_transfer, bool):
            raise TypeError(f"Bad end of transfer flag: {type(self.end_of_transfer).__name__}")

        if not isinstance(self.payload, memoryview):
            raise TypeError(f"Bad payload type: {type(self.payload).__name__}")

    @property
    def single_frame_transfer(self) -> bool:
        return self.index == 0 and self.end_of_transfer

    def __repr__(self) -> str:
        """
        If the payload is unreasonably long for a sensible string representation,
        it is truncated and suffixed with an ellipsis.
        """
        payload_length_limit = 100
        if len(self.payload) > payload_length_limit:
            payload = bytes(self.payload[:payload_length_limit]).hex() + "..."
        else:
            payload = bytes(self.payload).hex()
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwargs["priority"] = self.priority.name
        kwargs["payload"] = payload
        return pyuavcan.util.repr_attributes(self, **kwargs)


# noinspection PyTypeChecker
def _unittest_frame_base_ctor() -> None:
    from pytest import raises
    from pyuavcan.transport import Priority

    Frame(priority=Priority.LOW, transfer_id=1234, index=321, end_of_transfer=True, payload=memoryview(b""))

    with raises(TypeError):
        Frame(priority=2, transfer_id=1234, index=321, end_of_transfer=True, payload=memoryview(b""))  # type: ignore

    with raises(TypeError):
        Frame(
            priority=Priority.LOW,
            transfer_id=1234,
            index=321,
            end_of_transfer=1,  # type: ignore
            payload=memoryview(b""),
        )

    with raises(TypeError):
        Frame(priority=Priority.LOW, transfer_id=1234, index=321, end_of_transfer=False, payload=b"")  # type: ignore

    with raises(ValueError):
        Frame(priority=Priority.LOW, transfer_id=-1, index=321, end_of_transfer=True, payload=memoryview(b""))

    with raises(ValueError):
        Frame(priority=Priority.LOW, transfer_id=0, index=-1, end_of_transfer=True, payload=memoryview(b""))



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import dataclasses
import pyuavcan


@dataclasses.dataclass(frozen=True)
class Capture:
    """
    This is the abstract data class for all events reported via the capture API.

    If a transport implementation defines multiple event types, it is recommended to define a common superclass
    for them such that it is always possible to determine which transport an event has arrived from using a single
    instance check.
    """

    timestamp: pyuavcan.transport.Timestamp

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.Transport]:
        """
        Static reference to the type of transport that can emit captures of this type.
        For example, for UAVCAN/serial it would be :class:`pyuavcan.transport.serial.SerialTransport`.
        Although the method is static, it shall be overridden by all inheritors.
        """
        raise NotImplementedError


CaptureCallback = typing.Callable[[Capture], None]


@dataclasses.dataclass(frozen=True)
class AlienSessionSpecifier:
    """
    See :class:`AlienTransfer` and the abstract transport model.
    """

    source_node_id: typing.Optional[int]
    """None represents an anonymous transfer."""

    destination_node_id: typing.Optional[int]
    """None represents a broadcast transfer."""

    data_specifier: pyuavcan.transport.DataSpecifier

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, self.data_specifier, source_node_id=self.source_node_id, destination_node_id=self.destination_node_id
        )


@dataclasses.dataclass(frozen=True)
class AlienTransferMetadata:
    priority: pyuavcan.transport.Priority

    transfer_id: int
    """
    For outgoing transfers over transports with cyclic transfer-ID the modulo is computed automatically.
    The user does not have to bother; although, if it is desired to match the spoofed transfer with some
    follow-up activity (like a service response), the user needs to compute the modulo manually for obvious reasons.
    """

    session_specifier: AlienSessionSpecifier

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, self.session_specifier, priority=self.priority.name, transfer_id=self.transfer_id
        )


@dataclasses.dataclass(frozen=True)
class AlienTransfer:
    """
    This type models a captured (sniffed) decoded transfer exchanged between a local node and a remote node,
    between *remote nodes*, misaddressed transfer, or a spoofed transfer.

    It is different from :class:`pyuavcan.transport.Transfer` because the latter is intended for normal communication,
    whereas this type is designed for advanced network diagnostics, which is a very different use case.
    You may notice that the regular transfer model does not include some information such as, say, the route specifier,
    because the respective behaviors are managed by the transport configuration.
    """

    metadata: AlienTransferMetadata

    fragmented_payload: pyuavcan.transport.FragmentedPayload
    """
    For reconstructed transfers the number of fragments equals the number of frames in the transfer.
    For outgoing transfers the number of fragments may be arbitrary, the payload is always rearranged correctly.
    """

    def __eq__(self, other: object) -> bool:
        """
        Transfers whose payload is fragmented differently but content-wise is identical compare equal.

        >>> from pyuavcan.transport import MessageDataSpecifier, Priority
        >>> meta = AlienTransferMetadata(Priority.LOW, 999, AlienSessionSpecifier(123, None, MessageDataSpecifier(888)))
        >>> a =  AlienTransfer(meta, fragmented_payload=[memoryview(b'abc'), memoryview(b'def')])
        >>> a == AlienTransfer(meta, fragmented_payload=[memoryview(b'abcd'), memoryview(b''), memoryview(b'ef')])
        True
        >>> a == AlienTransfer(meta, fragmented_payload=[memoryview(b'abcdef')])
        True
        >>> a == AlienTransfer(meta, fragmented_payload=[])
        False
        """
        if isinstance(other, AlienTransfer):

            def cat(fp: pyuavcan.transport.FragmentedPayload) -> memoryview:
                return fp[0] if len(fp) == 1 else memoryview(b"".join(fp))

            return self.metadata == other.metadata and cat(self.fragmented_payload) == cat(other.fragmented_payload)
        return NotImplemented

    def __repr__(self) -> str:
        fragmented_payload = "+".join(f"{len(x)}B" for x in self.fragmented_payload)
        return pyuavcan.util.repr_attributes(self, self.metadata, fragmented_payload=f"[{fragmented_payload}]")


@dataclasses.dataclass(frozen=True)
class Trace:
    """
    Base event reconstructed by :class:`Tracer`.
    Transport-specific implementations may define custom subclasses.
    """

    timestamp: pyuavcan.transport.Timestamp
    """
    The local time when the traced event took place or was commenced.
    For transfers, this is the timestamp of the first frame.
    """


@dataclasses.dataclass(frozen=True)
class ErrorTrace(Trace):
    """
    This trace is yielded when the tracer has determined that it is unable to reconstruct a transfer.
    It may be further specialized by transport implementations.
    """


@dataclasses.dataclass(frozen=True)
class TransferTrace(Trace):
    """
    Reconstructed network data transfer (possibly exchanged between remote nodes) along with metadata.
    """

    transfer: AlienTransfer

    transfer_id_timeout: float
    """
    The tracer uses heuristics to automatically deduce the optimal transfer-ID timeout value per session
    based on the supplied captures.
    Whenever a new transfer is reassembled, the auto-deduced transfer-ID timeout that is currently used
    for its session is reported for informational purposes.
    This value may be used later to perform transfer deduplication if redundant tracers are used;
    for that, see :mod:`pyuavcan.transport.redundant`.
    """


class Tracer(abc.ABC):
    """
    The tracer takes single instances of :class:`Capture` at the input and delivers a reconstructed high-level
    view of network events (modeled by :class:`Trace`) at the output.
    It keeps massive internal state that is modified whenever :meth:`update` is invoked.
    The class may be used either for real-time analysis on a live network, or for post-mortem analysis with capture
    events read from a black box recorder or a log file.

    Instances of this class are entirely isolated from the outside world; they do not perform any IO and do not hold
    any resources, they are purely computing entities.
    To reset the state (e.g., in order to start analyzing a new log) simply discard the old instance and use a new one.

    The user should never attempt to instantiate implementations manually; instead, the factory method
    :meth:`pyuavcan.transport.Transport.make_tracer` should be used.

    Each transport implementation typically implements its own tracer.
    """

    @abc.abstractmethod
    def update(self, cap: Capture) -> typing.Optional[Trace]:
        """
        Takes a captured low-level network event at the input, returns a reconstructed high-level event at the output.
        If the event is considered irrelevant or did not update the internal state significantly
        (i.e., this is a non-last frame of a multi-frame transfer), the output is None.
        Reconstructed multi-frame transfers are reported as a single event when the last frame is received.

        Capture instances that are not supported by the current transport are silently ignored and None is returned.
        """
        raise NotImplementedError



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
from typing import Iterator, Optional, Sequence, Callable
import itertools
import pyuavcan
from .register import ValueProxy, Natural16, Natural32, RelaxedValue

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:  # pragma: no cover
    from typing import MutableMapping  # pylint: disable=ungrouped-imports


def make_transport(
    registers: MutableMapping[str, ValueProxy],
    *,
    reconfigurable: bool = False,
) -> Optional[pyuavcan.transport.Transport]:
    """
    Constructs a transport instance based on the configuration encoded in the supplied registers.
    If more than one transport is defined, a redundant instance will be constructed.

    The register schema is documented below per transport class
    (refer to the transport class documentation to find the defaults for optional registers).
    All transports also accept the following standard regsiters:

    +-------------------+-------------------+-----------------------------------------------------------------------+
    | Register name     | Register type     | Semantics                                                             |
    +===================+===================+=======================================================================+
    | ``uavcan.node.id``| ``natural16[1]``  | The node-ID to use. If the value exceeds the valid                    |
    |                   |                   | range, the constructed node will be anonymous.                        |
    +-------------------+-------------------+-----------------------------------------------------------------------+

    ..  list-table:: :mod:`pyuavcan.transport.udp`
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.udp.iface``
          - ``string``
          - Whitespace-separated list of /16 IP subnet addresses.
            16 least significant bits are replaced with the node-ID if configured, otherwise left unchanged.
            E.g.: ``127.42.0.42``: node-ID 257, result ``127.42.1.1``;
            ``127.42.0.42``: anonymous, result ``127.42.0.42``.

        * - ``uavcan.udp.duplicate_service_transfers``
          - ``bit[1]``
          - Apply deterministic data loss mitigation to RPC-service transfers by setting multiplication factor = 2.

        * - ``uavcan.udp.mtu``
          - ``natural16[1]``
          - The MTU for all constructed transport instances.

    ..  list-table:: :mod:`pyuavcan.transport.serial`
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.serial.iface``
          - ``string``
          - Whitespace-separated list of serial port names.
            E.g.: ``/dev/ttyACM0``, ``COM9``, ``socket://127.0.0.1:50905``.

        * - ``uavcan.serial.duplicate_service_transfers``
          - ``bit[1]``
          - Apply deterministic data loss mitigation to RPC-service transfers by setting multiplication factor = 2.

        * - ``uavcan.serial.baudrate``
          - ``natural16[1]``
          - The baudrate to set for all specified serial ports. Leave unchanged if zero.

    ..  list-table:: :mod:`pyuavcan.transport.can`
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.can.iface``
          - ``string``
          - Whitespace-separated list of CAN iface names.
            Each iface name shall follow the format defined in :class:`pyuavcan.transport.can.media.pythoncan`.
            E.g.: ``socketcan:vcan0``.

        * - ``uavcan.can.mtu``
          - ``natural16[1]``
          - The MTU value to use with all constructed CAN transports.
            Values other than 8 and 64 should not be used.

        * - ``uavcan.can.bitrate``
          - ``natural32[2]``
          - The bitrates to use for all constructed CAN transports
            for arbitration (first value) and data (second value) segments.
            To use Classic CAN, set both to the same value and set MTU = 8.

    ..  list-table:: :mod:`pyuavcan.transport.loopback`
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.loopback``
          - ``bit[1]``
          - If True, a loopback transport will be constructed. This is intended for testing only.

    :param registers:
        A mutable mapping of :class:`str` to :class:`pyuavcan.application.register.ValueProxy`.
        Normally, it should be constructed by :func:`pyuavcan.application.make_registry`.

    :param reconfigurable:
        If False (default), the return value is:

        - None if the registers do not encode a valid transport configuration.
        - A single transport instance if a non-redundant configuration is defined.
        - An instance of :class:`pyuavcan.transport.RedundantTransport` if more than one transport
          configuration is defined.

        If True, then the returned instance is always of type :class:`pyuavcan.transport.RedundantTransport`,
        where the set of inferiors is empty if no transport configuration is defined.
        This case is intended for applications that may want to change the transport configuration afterwards.

    :return:
        None if no transport is configured AND ``reconfigurable`` is False.
        Otherwise, a functional transport instance is returned.

    :raises:
        - :class:`pyuavcan.application.register.MissingRegisterError` if a register is expected but cannot be found.
        - :class:`pyuavcan.application.register.ValueConversionError` if a register is found but its value
          cannot be converted to the correct type.

    >>> from pyuavcan.application.register import ValueProxy, Natural16, Natural32
    >>> reg = {
    ...     "uavcan.udp.iface": ValueProxy("127.99.0.0"),
    ...     "uavcan.node.id": ValueProxy(Natural16([257])),
    ... }
    >>> tr = make_transport(reg)
    >>> tr
    UDPTransport('127.99.1.1', local_node_id=257, ...)
    >>> tr.close()
    >>> tr = make_transport(reg, reconfigurable=True)                   # Same but reconfigurable.
    >>> tr                                                              # Wrapped into RedundantTransport.
    RedundantTransport(UDPTransport('127.99.1.1', local_node_id=257, ...))
    >>> tr.close()

    >>> int(reg["uavcan.udp.mtu"])      # Defaults created automatically to expose all configurables.
    1200
    >>> int(reg["uavcan.can.mtu"])
    64
    >>> reg["uavcan.can.bitrate"].ints
    [1000000, 4000000]

    >>> reg = {                                             # Triply-redundant heterogeneous transport:
    ...     "uavcan.udp.iface":    ValueProxy("127.99.0.15 127.111.0.15"),  # Double UDP transport
    ...     "uavcan.serial.iface": ValueProxy("socket://127.0.0.1:50905"),  # Serial transport
    ... }
    >>> tr = make_transport(reg)                            # The node-ID was not set, so the transport is anonymous.
    >>> tr                                          # doctest: +NORMALIZE_WHITESPACE
    RedundantTransport(UDPTransport('127.99.0.15',  local_node_id=None, ...),
                       UDPTransport('127.111.0.15', local_node_id=None, ...),
                       SerialTransport('socket://127.0.0.1:50905', local_node_id=None, ...))
    >>> tr.close()

    >>> reg = {
    ...     "uavcan.can.iface":   ValueProxy("virtual: virtual:"),    # Doubly-redundant CAN
    ...     "uavcan.can.mtu":     ValueProxy(Natural16([32])),
    ...     "uavcan.can.bitrate": ValueProxy(Natural32([500_000, 2_000_000])),
    ...     "uavcan.node.id":     ValueProxy(Natural16([123])),
    ... }
    >>> tr = make_transport(reg)
    >>> tr                                          # doctest: +NORMALIZE_WHITESPACE
    RedundantTransport(CANTransport(PythonCANMedia('virtual:', mtu=32), local_node_id=123),
                       CANTransport(PythonCANMedia('virtual:', mtu=32), local_node_id=123))
    >>> tr.close()

    >>> reg = {
    ...     "uavcan.udp.iface": ValueProxy("127.99.1.1"),       # Per the standard register specs,
    ...     "uavcan.node.id": ValueProxy(Natural16([0xFFFF])),  # 0xFFFF means unset/anonymous.
    ... }
    >>> tr = make_transport(reg)
    >>> tr
    UDPTransport('127.99.1.1', local_node_id=None, ...)
    >>> tr.close()

    >>> tr = make_transport({})
    >>> tr is None
    True
    >>> tr = make_transport({}, reconfigurable=True)
    >>> tr                                                          # Redundant transport with no inferiors.
    RedundantTransport()
    """

    def init(name: str, default: RelaxedValue) -> ValueProxy:
        return registers.setdefault("uavcan." + name, ValueProxy(default))

    # Per Specification, if uavcan.node.id = 65535, the node-ID is unspecified.
    node_id: Optional[int] = int(init("node.id", Natural16([0xFFFF])))
    # TODO: currently, we raise an error if the node-ID setting exceeds the maximum allowed value for the current
    # transport, but the spec recommends that we should handle this as if the node-ID was not set at all.
    if node_id is not None and not (0 <= node_id < 0xFFFF):
        node_id = None

    transports = list(itertools.chain(*(f(registers, node_id) for f in _SPECIALIZATIONS)))
    assert all(isinstance(t, pyuavcan.transport.Transport) for t in transports)

    if not reconfigurable:
        if not transports:
            return None
        if len(transports) == 1:
            return transports[0]

    from pyuavcan.transport.redundant import RedundantTransport

    red = RedundantTransport()
    for tr in transports:
        red.attach_inferior(tr)
    return red


def _make_udp(
    registers: MutableMapping[str, ValueProxy], node_id: Optional[int]
) -> Iterator[pyuavcan.transport.Transport]:
    def init(name: str, default: RelaxedValue) -> ValueProxy:
        return registers.setdefault("uavcan.udp." + name, ValueProxy(default))

    ip_list = str(init("iface", "")).split()
    mtu = int(init("mtu", Natural16([1200])))
    srv_mult = int(init("duplicate_service_transfers", False)) + 1

    if ip_list:
        from pyuavcan.transport.udp import UDPTransport

        for ip in ip_list:
            yield UDPTransport(ip, node_id, mtu=mtu, service_transfer_multiplier=srv_mult)


def _make_serial(
    registers: MutableMapping[str, ValueProxy], node_id: Optional[int]
) -> Iterator[pyuavcan.transport.Transport]:
    def init(name: str, default: RelaxedValue) -> ValueProxy:
        return registers.setdefault("uavcan.serial." + name, ValueProxy(default))

    port_list = str(init("iface", "")).split()
    srv_mult = int(init("duplicate_service_transfers", False)) + 1
    baudrate = int(init("baudrate", Natural32([0]))) or None

    if port_list:
        from pyuavcan.transport.serial import SerialTransport

        for port in port_list:
            yield SerialTransport(str(port), node_id, service_transfer_multiplier=srv_mult, baudrate=baudrate)


def _make_can(
    registers: MutableMapping[str, ValueProxy], node_id: Optional[int]
) -> Iterator[pyuavcan.transport.Transport]:
    def init(name: str, default: RelaxedValue) -> ValueProxy:
        return registers.setdefault("uavcan.can." + name, ValueProxy(default))

    iface_list = str(init("iface", "")).split()
    mtu = int(init("mtu", Natural16([64])))
    br_arb, br_data = init("bitrate", Natural32([1_000_000, 4_000_000])).ints

    if iface_list:
        from pyuavcan.transport.can import CANTransport

        for iface in iface_list:
            media: pyuavcan.transport.can.media.Media
            if iface.lower().startswith("socketcan:"):
                from pyuavcan.transport.can.media.socketcan import SocketCANMedia

                media = SocketCANMedia(iface.split(":")[-1], mtu=mtu)
            else:
                from pyuavcan.transport.can.media.pythoncan import PythonCANMedia

                media = PythonCANMedia(iface, br_arb if br_arb == br_data else (br_arb, br_data), mtu)
            yield CANTransport(media, node_id)


def _make_loopback(
    registers: MutableMapping[str, ValueProxy], node_id: Optional[int]
) -> Iterator[pyuavcan.transport.Transport]:
    # Not sure if exposing this is a good idea because the loopback transport is hardly useful outside of test envs.
    if registers.setdefault("uavcan.loopback", ValueProxy(False)):
        from pyuavcan.transport.loopback import LoopbackTransport

        yield LoopbackTransport(node_id)


_SPECIALIZATIONS: Sequence[
    Callable[[MutableMapping[str, ValueProxy], Optional[int]], Iterator[pyuavcan.transport.Transport]]
] = [v for k, v in globals().items() if callable(v) and k.startswith("_make_")]
assert len(_SPECIALIZATIONS) >= 4



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This module contains implementations of various CRC algorithms used by the transports.

`32-Bit Cyclic Redundancy Codes for Internet Applications (Philip Koopman)
<https://users.ece.cmu.edu/~koopman/networks/dsn02/dsn02_koopman.pdf>`_.
"""

from ._base import CRCAlgorithm as CRCAlgorithm
from ._crc16_ccitt import CRC16CCITT as CRC16CCITT
from ._crc32c import CRC32C as CRC32C
from ._crc64we import CRC64WE as CRC64WE



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._socket_factory import SocketFactory as SocketFactory
from ._socket_factory import Sniffer as Sniffer

from ._endpoint_mapping import IP_ADDRESS_NODE_ID_MASK as IP_ADDRESS_NODE_ID_MASK
from ._endpoint_mapping import MULTICAST_GROUP_SUBJECT_ID_MASK as MULTICAST_GROUP_SUBJECT_ID_MASK
from ._endpoint_mapping import SUBJECT_PORT as SUBJECT_PORT
from ._endpoint_mapping import node_id_to_unicast_ip as node_id_to_unicast_ip
from ._endpoint_mapping import unicast_ip_to_node_id as unicast_ip_to_node_id
from ._endpoint_mapping import message_data_specifier_to_multicast_group as message_data_specifier_to_multicast_group
from ._endpoint_mapping import multicast_group_to_message_data_specifier as multicast_group_to_message_data_specifier
from ._endpoint_mapping import service_data_specifier_to_udp_port as service_data_specifier_to_udp_port
from ._endpoint_mapping import udp_port_to_service_data_specifier as udp_port_to_service_data_specifier

from ._link_layer import LinkLayerPacket as LinkLayerPacket
from ._link_layer import LinkLayerCapture as LinkLayerCapture



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import os
import errno
import shutil
import typing
import pathlib
import logging
import itertools
import numpy as np
import pyuavcan
import pyuavcan.application
from uavcan.file import Path_2_0 as Path
from uavcan.file import Error_1_0 as Error
from uavcan.file import Read_1_1 as Read
from uavcan.file import Write_1_1 as Write
from uavcan.file import List_0_2 as List
from uavcan.file import GetInfo_0_2 as GetInfo
from uavcan.file import Modify_1_1 as Modify
from uavcan.primitive import Unstructured_1_0 as Unstructured


class FileServer:
    """
    Exposes local filesystems via the standard RPC-services defined in ``uavcan.file``.
    The lifetime of this instance matches the lifetime of its node.
    """

    def __init__(
        self, node: pyuavcan.application.Node, roots: typing.Iterable[typing.Union[str, pathlib.Path]]
    ) -> None:
        """
        :param node:
            The node instance to initialize the file server on.
            It shall not be anonymous, otherwise it's a
            :class:`pyuavcan.transport.OperationNotDefinedForAnonymousNodeError`.

        :param roots:
            All file operations will be performed in the specified directories.
            The first directory to match takes precedence.
            New files are created in the first directory.
        """
        self._roots = [pathlib.Path(x).resolve() for x in roots]

        # noinspection PyUnresolvedReferences
        self._data_transfer_capacity = int(pyuavcan.dsdl.get_model(Unstructured)["value"].data_type.capacity)

        s_ls = node.get_server(List)
        s_if = node.get_server(GetInfo)
        s_mo = node.get_server(Modify)
        s_rd = node.get_server(Read)
        s_wr = node.get_server(Write)

        def start() -> None:
            s_ls.serve_in_background(self._serve_ls)
            s_if.serve_in_background(self._serve_if)
            s_mo.serve_in_background(self._serve_mo)
            s_rd.serve_in_background(self._serve_rd)
            s_wr.serve_in_background(self._serve_wr)

        def close() -> None:
            s_ls.close()
            s_if.close()
            s_mo.close()
            s_rd.close()
            s_wr.close()

        node.add_lifetime_hooks(start, close)

    @property
    def roots(self) -> typing.List[pathlib.Path]:
        """
        File operations will be performed within these root directories.
        The first directory to match takes precedence.
        New files are created in the first directory in the list.
        The list can be modified.
        """
        return self._roots

    def locate(self, p: typing.Union[pathlib.Path, str, Path]) -> typing.Tuple[pathlib.Path, pathlib.Path]:
        """
        Iterate through :attr:`roots` until a root r is found such that ``r/p`` exists and return ``(r, p)``.
        Otherwise, return nonexistent ``(roots[0], p)``.
        The leading slash makes no difference because we only search through the specified roots.

        :raises: :class:`FileNotFoundError` if :attr:`roots` is empty.
        """
        if isinstance(p, Path):
            p = p.path.tobytes().decode(errors="ignore").replace(chr(Path.SEPARATOR), os.sep)
        assert not isinstance(p, Path)
        p = pathlib.Path(str(pathlib.Path(p)).strip(os.sep))  # Make relative, canonicalize the trailing separator
        # See if there are existing entries under this name:
        for r in self.roots:
            if (r / p).exists():
                return r, p
        # If not, assume that we are going to create one:
        if len(self.roots) > 0:
            return self.roots[0], p
        raise FileNotFoundError(str(p))

    def glob(self, pat: str) -> typing.Iterable[typing.Tuple[pathlib.Path, pathlib.Path]]:
        """
        Search for entries matching the pattern across :attr:`roots`, in order.
        Return tuple of (root, match), where match is relative to its root.
        Ordering not enforced.
        """
        pat = pat.strip(os.sep)
        for d in self.roots:
            for e in d.glob(pat):
                yield d, e.absolute().relative_to(d.absolute())

    @staticmethod
    def convert_error(ex: Exception) -> Error:
        for ty, err in {
            FileNotFoundError: Error.NOT_FOUND,
            IsADirectoryError: Error.IS_DIRECTORY,
            NotADirectoryError: Error.NOT_SUPPORTED,
            PermissionError: Error.ACCESS_DENIED,
            FileExistsError: Error.INVALID_VALUE,
        }.items():
            if isinstance(ex, ty):
                return Error(err)
        if isinstance(ex, OSError):
            return Error(
                {
                    errno.EACCES: Error.ACCESS_DENIED,
                    errno.E2BIG: Error.FILE_TOO_LARGE,
                    errno.EINVAL: Error.INVALID_VALUE,
                    errno.EIO: Error.IO_ERROR,
                    errno.EISDIR: Error.IS_DIRECTORY,
                    errno.ENOENT: Error.NOT_FOUND,
                    errno.ENOTSUP: Error.NOT_SUPPORTED,
                    errno.ENOSPC: Error.OUT_OF_SPACE,
                }.get(ex.errno, Error.UNKNOWN_ERROR)
            )
        return Error(Error.UNKNOWN_ERROR)

    async def _serve_ls(
        self, request: List.Request, meta: pyuavcan.presentation.ServiceRequestMetadata
    ) -> List.Response:
        _logger.info("%r: Request from %r: %r", self, meta.client_node_id, request)
        try:
            d = pathlib.Path(*self.locate(request.directory_path))
            for i, e in enumerate(sorted(d.iterdir())):
                if i == request.entry_index:
                    rel = e.absolute().relative_to(d.absolute())
                    return List.Response(Path(str(rel)))
        except FileNotFoundError:
            pass
        except Exception as ex:
            _logger.exception("%r: Directory list error: %s", self, ex)
        return List.Response()

    async def _serve_if(
        self, request: GetInfo.Request, meta: pyuavcan.presentation.ServiceRequestMetadata
    ) -> GetInfo.Response:
        _logger.info("%r: Request from %r: %r", self, meta.client_node_id, request)
        try:
            p = pathlib.Path(*self.locate(request.path))
            return GetInfo.Response(
                size=p.resolve().stat().st_size,
                unix_timestamp_of_last_modification=int(p.resolve().stat().st_mtime),
                is_file_not_directory=p.is_file() or not p.is_dir(),  # Handle special files like /dev/null correctly
                is_link=os.path.islink(p),
                is_readable=os.access(p, os.R_OK),
                is_writeable=os.access(p, os.W_OK),
            )
        except Exception as ex:
            _logger.info("%r: Error: %r", self, ex, exc_info=True)
            return GetInfo.Response(self.convert_error(ex))

    async def _serve_mo(
        self, request: Modify.Request, meta: pyuavcan.presentation.ServiceRequestMetadata
    ) -> Modify.Response:
        _logger.info("%r: Request from %r: %r", self, meta.client_node_id, request)

        try:
            if len(request.destination.path) == 0:  # No destination: remove
                p = pathlib.Path(*self.locate(request.source))
                if p.is_dir():
                    shutil.rmtree(p)
                else:
                    p.unlink()
                return Modify.Response()

            if len(request.source.path) == 0:  # No source: touch
                dst = pathlib.Path(*self.locate(request.destination)).resolve()
                dst.parent.mkdir(parents=True, exist_ok=True)
                dst.touch(exist_ok=True)
                return Modify.Response()

            # Resolve paths and ensure the target directory exists.
            src = pathlib.Path(*self.locate(request.source)).resolve()
            dst = pathlib.Path(*self.locate(request.destination)).resolve()
            dst.parent.mkdir(parents=True, exist_ok=True)

            # At this point if src does not exist it is definitely an error.
            if not src.exists():
                return Modify.Response(Error(Error.NOT_FOUND))

            # Can't proceed if destination exists but overwrite is not enabled.
            if dst.exists():
                if not request.overwrite_destination:
                    return Modify.Response(Error(Error.INVALID_VALUE))
                if dst.is_dir():
                    shutil.rmtree(dst, ignore_errors=True)
                else:
                    dst.unlink()

            # Do move/copy depending on the flag.
            if request.preserve_source:
                if src.is_dir():
                    shutil.copytree(src, dst)
                else:
                    shutil.copy(src, dst)
            else:
                shutil.move(str(src), str(dst))
            return Modify.Response()
        except Exception as ex:
            _logger.info("%r: Error: %r", self, ex, exc_info=True)
            return Modify.Response(self.convert_error(ex))

    async def _serve_rd(
        self, request: Read.Request, meta: pyuavcan.presentation.ServiceRequestMetadata
    ) -> Read.Response:
        _logger.info("%r: Request from %r: %r", self, meta.client_node_id, request)
        try:
            with open(pathlib.Path(*self.locate(request.path)), "rb") as f:
                if request.offset != 0:  # Do not seek unless necessary to support non-seekable files.
                    f.seek(request.offset)
                data = f.read(self._data_transfer_capacity)
            return Read.Response(data=Unstructured(np.frombuffer(data, np.uint8)))
        except Exception as ex:
            _logger.info("%r: Error: %r", self, ex, exc_info=True)
            return Read.Response(self.convert_error(ex))

    async def _serve_wr(
        self, request: Write.Request, meta: pyuavcan.presentation.ServiceRequestMetadata
    ) -> Write.Response:
        _logger.info("%r: Request from %r: %r", self, meta.client_node_id, request)
        try:
            data = request.data.value.tobytes()
            with open(pathlib.Path(*self.locate(request.path)), "rb+") as f:
                f.seek(request.offset)
                f.write(data)
                if not data:
                    f.truncate()
            return Write.Response()
        except Exception as ex:
            _logger.info("%r: Error: %r", self, ex, exc_info=True)
            return Write.Response(self.convert_error(ex))

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, list(map(str, self.roots)))


class FileClient:
    """
    A trivial proxy that provides a higher-level and more pythonic API on top of the standard RPC-services
    from ``uavcan.file``.
    Client instances are created lazily at first request and then kept alive until this instance is closed.
    All remote operations raise :class:`FileTimeoutError` on timeout.
    """

    def __init__(
        self,
        local_node: pyuavcan.application.Node,
        server_node_id: int,
        response_timeout: float = 3.0,
        priority: pyuavcan.transport.Priority = pyuavcan.transport.Priority.SLOW,
    ) -> None:
        """
        :param local_node: RPC-service clients will be created on this node.
        :param server_node_id: All requests will be sent to this node-ID.
        :param response_timeout: Raise :class:`FileTimeoutError` if the server does not respond in this time.
        :param priority: Transfer priority for requests (and, therefore, responses).
        """
        self._node = local_node
        self._server_node_id = server_node_id
        self._response_timeout = float(response_timeout)
        # noinspection PyArgumentList
        self._priority = pyuavcan.transport.Priority(priority)

        self._clients: typing.Dict[
            typing.Type[pyuavcan.dsdl.ServiceObject],
            pyuavcan.presentation.Client[pyuavcan.dsdl.ServiceObject],
        ] = {}

        # noinspection PyUnresolvedReferences
        self._data_transfer_capacity = int(pyuavcan.dsdl.get_model(Unstructured)["value"].data_type.capacity)

    @property
    def data_transfer_capacity(self) -> int:
        """
        A convenience constant derived from DSDL: the maximum number of bytes per read/write transfer.
        Larger reads/writes are non-atomic.
        """
        return self._data_transfer_capacity

    @property
    def server_node_id(self) -> int:
        """
        The node-ID of the remote file server.
        """
        return self._server_node_id

    def close(self) -> None:
        """
        Close all RPC-service client instances created up to this point.
        """
        for c in self._clients.values():
            c.close()
        self._clients.clear()

    async def list(self, path: str) -> typing.AsyncIterable[str]:
        """
        Proxy for ``uavcan.file.List``. Invokes requests in series until all elements are listed.
        """
        for index in itertools.count():
            res = await self._call(List, List.Request(entry_index=index, directory_path=Path(path)))
            assert isinstance(res, List.Response)
            p = res.entry_base_name.path.tobytes().decode(errors="ignore")
            if p:
                yield str(p)
            else:
                break

    async def get_info(self, path: str) -> GetInfo.Response:
        """
        Proxy for ``uavcan.file.GetInfo``. Be sure to check the error code in the returned object.
        """
        res = await self._call(GetInfo, GetInfo.Request(Path(path)))
        assert isinstance(res, GetInfo.Response)
        return res

    async def remove(self, path: str) -> int:
        """
        Proxy for ``uavcan.file.Modify``.

        :returns: See ``uavcan.file.Error``
        """
        res = await self._call(Modify, Modify.Request(source=Path(path)))
        assert isinstance(res, Modify.Response)
        return int(res.error.value)

    async def touch(self, path: str) -> int:
        """
        Proxy for ``uavcan.file.Modify``.

        :returns: See ``uavcan.file.Error``
        """
        res = await self._call(Modify, Modify.Request(destination=Path(path)))
        assert isinstance(res, Modify.Response)
        return int(res.error.value)

    async def copy(self, src: str, dst: str, overwrite: bool = False) -> int:
        """
        Proxy for ``uavcan.file.Modify``.

        :returns: See ``uavcan.file.Error``
        """
        res = await self._call(
            Modify,
            Modify.Request(
                preserve_source=True,
                overwrite_destination=overwrite,
                source=Path(src),
                destination=Path(dst),
            ),
        )
        assert isinstance(res, Modify.Response)
        return int(res.error.value)

    async def move(self, src: str, dst: str, overwrite: bool = False) -> int:
        """
        Proxy for ``uavcan.file.Modify``.

        :returns: See ``uavcan.file.Error``
        """
        res = await self._call(
            Modify,
            Modify.Request(
                preserve_source=False,
                overwrite_destination=overwrite,
                source=Path(src),
                destination=Path(dst),
            ),
        )
        assert isinstance(res, Modify.Response)
        return int(res.error.value)

    async def read(self, path: str, offset: int = 0, size: typing.Optional[int] = None) -> typing.Union[int, bytes]:
        """
        Proxy for ``uavcan.file.Read``.

        :param path:
            The file to read.

        :param offset:
            Read offset from the beginning of the file.
            Currently, it must be positive; negative offsets from the end of the file may be supported later.

        :param size:
            Read requests will be stopped after the end of the file is reached or at least this many bytes are read.
            If None (default), the entire file will be read (this may exhaust local memory).
            If zero, this call is a no-op.

        :returns:
            ``uavcan.file.Error.value`` on error (e.g., no file),
            data on success (empty if the offset is out of bounds or the file is empty).
        """

        async def once() -> typing.Union[int, bytes]:
            res = await self._call(Read, Read.Request(offset=offset, path=Path(path)))
            assert isinstance(res, Read.Response)
            if res.error.value != 0:
                return int(res.error.value)
            return bytes(res.data.value.tobytes())

        if size is None:
            size = 2 ** 64
        data = b""
        while len(data) < size:
            out = await once()
            if isinstance(out, int):
                return out
            assert isinstance(out, bytes)
            if not out:
                break
            data += out
            offset += len(out)
        return data

    async def write(
        self, path: str, data: typing.Union[memoryview, bytes], offset: int = 0, *, truncate: bool = True
    ) -> int:
        """
        Proxy for ``uavcan.file.Write``.

        :param path:
            The file to write.

        :param data:
            The data to write at the specified offset.
            The number of write requests depends on the size of data.

        :param offset:
            Write offset from the beginning of the file.
            Currently, it must be positive; negative offsets from the end of the file may be supported later.

        :param truncate:
            If True, the rest of the file after ``offset + len(data)`` will be truncated.
            This is done by sending an empty write request, as prescribed by the Specification.

        :returns: See ``uavcan.file.Error``
        """

        async def once(d: typing.Union[memoryview, bytes]) -> int:
            res = await self._call(
                Write, Write.Request(offset, path=Path(path), data=Unstructured(np.frombuffer(d, np.uint8)))
            )
            assert isinstance(res, Write.Response)
            return res.error.value

        limit = self.data_transfer_capacity
        while len(data) > 0:
            frag, data = data[:limit], data[limit:]
            out = await once(frag)
            offset += len(frag)
            if out != 0:
                return out
        if truncate:
            return await once(b"")
        return 0

    async def _call(
        self, ty: typing.Type[pyuavcan.dsdl.ServiceObject], request: pyuavcan.dsdl.CompositeObject
    ) -> pyuavcan.dsdl.CompositeObject:
        try:
            cln = self._clients[ty]
        except LookupError:
            self._clients[ty] = self._node.make_client(ty, self._server_node_id)
            cln = self._clients[ty]
            cln.response_timeout = self._response_timeout
            cln.priority = self._priority

        result = await cln.call(request)
        if result is None:
            raise FileTimeoutError(f"File service call timed out on {cln}")
        return result[0]

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self._node, server_node_id=self._server_node_id)


class FileTimeoutError(pyuavcan.application.NetworkTimeoutError):
    """
    The specialization of the network error for file access.
    """


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
from pyuavcan.transport import Timestamp
from ._frame import SerialFrame


class StreamParser:
    """
    A stream parser is fed with bytes received from the channel.
    The parser maintains internal parsing state machine; whenever the machine detects that a valid frame is received,
    the callback is invoked.

    When the state machine identifies that a received block of data cannot possibly
    contain or be part of a valid frame, the raw bytes are delivered into the callback as-is for optional later
    processing; such data is called "out-of-band" (OOB) data. An empty sequence of OOB bytes is never reported.
    The OOB data reporting can be useful if the same serial port is used both for UAVCAN and as a text console.
    The OOB bytes may or may not be altered by the COBS decoding logic.
    """

    def __init__(
        self,
        callback: typing.Callable[[Timestamp, memoryview, typing.Optional[SerialFrame]], None],
        max_payload_size_bytes: int,
    ):
        """
        :param callback: Invoked when a new frame is parsed or when a block of data could not be recognized as a frame.
            In the case of success, an instance of the frame class is passed in the last argument, otherwise it's None.
            In either case, the raw buffer is supplied as the second argument for capture/diagnostics or OOB handling.

        :param max_payload_size_bytes: Frames containing more than this many bytes of payload
            (after escaping and not including the header, CRC, and delimiters) may be considered invalid.
            This is to shield the parser against OOM errors when subjected to an invalid stream of bytes.
        """
        if not (callable(callback) and max_payload_size_bytes > 0):
            raise ValueError("Invalid parameters")

        self._callback = callback
        self._max_frame_size_bytes = (
            SerialFrame.calc_cobs_size(
                max_payload_size_bytes + SerialFrame.NUM_OVERHEAD_BYTES_EXCEPT_DELIMITERS_AND_ESCAPING
            )
            + 2
        )
        self._buffer = bytearray()  # Entire frame including all delimiters.
        self._timestamp: typing.Optional[Timestamp] = None

    def process_next_chunk(self, chunk: typing.Union[bytes, bytearray, memoryview], timestamp: Timestamp) -> None:
        # TODO: PERFORMANCE WARNING: DECODE COBS ON THE FLY TO AVOID EXTRA COPYING
        for b in chunk:
            self._buffer.append(b)
            if b == SerialFrame.FRAME_DELIMITER_BYTE:
                self._finalize(known_invalid=self._outside_frame)
            else:
                if self._timestamp is None:
                    self._timestamp = timestamp  # https://github.com/UAVCAN/pyuavcan/issues/112

        if self._outside_frame or (len(self._buffer) > self._max_frame_size_bytes):
            self._finalize(known_invalid=True)

    @property
    def _outside_frame(self) -> bool:
        return self._timestamp is None

    def _finalize(self, known_invalid: bool) -> None:
        if not self._buffer or (len(self._buffer) == 1 and self._buffer[0] == SerialFrame.FRAME_DELIMITER_BYTE):
            # Avoid noise in the OOB output during normal operation.
            # TODO: this is a hack in place of the proper on-the-fly COBS parser.
            return

        buf = memoryview(self._buffer)
        self._buffer = bytearray()  # There are memoryview instances pointing to the old buffer!
        ts = self._timestamp or Timestamp.now()
        self._timestamp = None

        parsed: typing.Optional[SerialFrame] = None
        if (not known_invalid) and len(buf) <= self._max_frame_size_bytes:
            parsed = SerialFrame.parse_from_cobs_image(buf)

        self._callback(ts, buf, parsed)


def _unittest_stream_parser() -> None:
    from pytest import raises
    from pyuavcan.transport import Priority, MessageDataSpecifier

    ts = Timestamp.now()

    outputs: typing.List[typing.Tuple[Timestamp, memoryview, typing.Optional[SerialFrame]]] = []

    with raises(ValueError):
        sp = StreamParser(lambda *_: None, 0)

    sp = StreamParser(lambda ts, buf, item: outputs.append((ts, buf, item)), 4)
    print("sp._max_frame_size_bytes:", sp._max_frame_size_bytes)  # pylint: disable=protected-access

    def proc(
        b: typing.Union[bytes, memoryview]
    ) -> typing.Sequence[typing.Tuple[Timestamp, memoryview, typing.Optional[SerialFrame]]]:
        sp.process_next_chunk(b, ts)
        out = outputs[:]
        outputs.clear()
        for i, (t, bb, f) in enumerate(out):
            print(f"output {i + 1} of {len(out)}: ", t, bytes(bb), f)
        return out

    assert not outputs
    ((tsa, buf, a),) = proc(b"abcdef\x00")
    assert ts.monotonic_ns <= tsa.monotonic_ns <= Timestamp.now().monotonic_ns
    assert ts.system_ns <= tsa.system_ns <= Timestamp.now().system_ns
    assert a is None
    assert memoryview(b"abcdef\x00") == buf
    assert [] == proc(b"")

    # Valid frame.
    f1 = SerialFrame(
        priority=Priority.HIGH,
        source_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        destination_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        data_specifier=MessageDataSpecifier(2345),
        transfer_id=1234567890123456789,
        index=1234567,
        end_of_transfer=True,
        payload=memoryview(b"ab\x9E\x8E"),
    )  # 4 bytes of payload.
    ((tsa, buf, a),) = proc(f1.compile_into(bytearray(100)))
    assert tsa == ts
    assert isinstance(a, SerialFrame)
    assert SerialFrame.__eq__(f1, a)
    assert buf[-1] == 0  # Frame delimiters are in place.

    # Second valid frame is too long.
    f2 = SerialFrame(
        priority=Priority.HIGH,
        source_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        destination_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        data_specifier=MessageDataSpecifier(2345),
        transfer_id=1234567890123456789,
        index=1234567,
        end_of_transfer=True,
        payload=memoryview(bytes(f1.compile_into(bytearray(1000))) * 2),
    )
    assert len(f2.payload) == 43 * 2  # Cobs escaping
    ((tsa, buf, a),) = proc(f2.compile_into(bytearray(1000)))
    assert tsa == ts
    assert a is None
    assert buf[-1] == 0  # Frame delimiters are in place.

    # Create new instance with much larger frame size limit; feed both frames but let the first one be incomplete.
    sp = StreamParser(lambda ts, buf, item: outputs.append((ts, buf, item)), 10 ** 6)
    assert [] == proc(f1.compile_into(bytearray(200))[:-2])  # First one is ended abruptly.
    (tsa, _, a), (tsb, _, b), = proc(
        f2.compile_into(bytearray(200))
    )  # Then the second frame begins.
    assert tsa == ts
    assert tsb == ts
    assert a is None
    assert isinstance(b, SerialFrame)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This module contains common classes and algorithms used in a certain category of transports
which we call **High Overhead Transports**.
They are designed for highly capable mediums where packets are large and data transfer speeds are high.

For example, UDP, Serial, and IEEE 802.15.4 are high-overhead transports.
CAN, on the other hand, is not a high-overhead transport;
none of the entities defined in this module can be used with CAN.
"""

from ._frame import Frame as Frame

from ._transfer_serializer import serialize_transfer as serialize_transfer

from ._transfer_reassembler import TransferReassembler as TransferReassembler

from ._common import TransferCRC as TransferCRC

from ._alien_transfer_reassembler import AlienTransferReassembler as AlienTransferReassembler



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import asyncio
import logging
import dataclasses
import pyuavcan
from pyuavcan.transport import Timestamp
from pyuavcan.transport.commons.high_overhead_transport import TransferReassembler
from .._frame import SerialFrame
from ._base import SerialSession


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SerialInputSessionStatistics(pyuavcan.transport.SessionStatistics):
    reassembly_errors_per_source_node_id: typing.Dict[
        int, typing.Dict[TransferReassembler.Error, int]
    ] = dataclasses.field(default_factory=dict)
    """
    Keys are source node-IDs; values are dicts where keys are error enum members and values are counts.
    """


class SerialInputSession(SerialSession, pyuavcan.transport.InputSession):
    DEFAULT_TRANSFER_ID_TIMEOUT = 2.0
    """
    Units are seconds. Can be overridden after instantiation if needed.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly.
        Instead, use the factory method :meth:`pyuavcan.transport.serial.SerialTransport.get_input_session`.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._loop = loop
        assert self._loop is not None

        self._statistics = SerialInputSessionStatistics()
        self._transfer_id_timeout = self.DEFAULT_TRANSFER_ID_TIMEOUT
        self._queue: asyncio.Queue[pyuavcan.transport.TransferFrom] = asyncio.Queue()
        self._reassemblers: typing.Dict[int, TransferReassembler] = {}

        super().__init__(finalizer)

    def _process_frame(self, timestamp: Timestamp, frame: SerialFrame) -> None:
        """
        This is a part of the transport-internal API. It's a public method despite the name because Python's
        visibility handling capabilities are limited. I guess we could define a private abstract base to
        handle this but it feels like too much work. Why can't we have protected visibility in Python?
        """
        assert frame.data_specifier == self._specifier.data_specifier, "Internal protocol violation"
        self._statistics.frames += 1

        transfer: typing.Optional[pyuavcan.transport.TransferFrom]
        if frame.source_node_id is None:
            transfer = TransferReassembler.construct_anonymous_transfer(timestamp, frame)
            if transfer is None:
                self._statistics.errors += 1
                _logger.debug("%s: Invalid anonymous frame: %s", self, frame)
        else:
            transfer = self._get_reassembler(frame.source_node_id).process_frame(
                timestamp, frame, self._transfer_id_timeout
            )
        if transfer is not None:
            self._statistics.transfers += 1
            self._statistics.payload_bytes += sum(map(len, transfer.fragmented_payload))
            _logger.debug("%s: Received transfer: %s; current stats: %s", self, transfer, self._statistics)
            try:
                self._queue.put_nowait(transfer)
            except asyncio.QueueFull:  # pragma: no cover
                # TODO: make the queue capacity configurable
                self._statistics.drops += len(transfer.fragmented_payload)

    async def receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        try:
            timeout = monotonic_deadline - self._loop.time()
            if timeout > 0:
                transfer = await asyncio.wait_for(self._queue.get(), timeout)
            else:
                transfer = self._queue.get_nowait()
        except (asyncio.TimeoutError, asyncio.QueueEmpty):
            # If there are unprocessed transfers, allow the caller to read them even if the instance is closed.
            self._raise_if_closed()
            return None
        else:
            assert isinstance(transfer, pyuavcan.transport.TransferFrom), "Internal protocol violation"
            assert transfer.source_node_id == self._specifier.remote_node_id or self._specifier.remote_node_id is None
            return transfer

    @property
    def transfer_id_timeout(self) -> float:
        return self._transfer_id_timeout

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        if value > 0:
            self._transfer_id_timeout = float(value)
        else:
            raise ValueError(f"Invalid value for transfer-ID timeout [second]: {value}")

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> SerialInputSessionStatistics:
        return copy.copy(self._statistics)

    def _get_reassembler(self, source_node_id: int) -> TransferReassembler:
        try:
            return self._reassemblers[source_node_id]
        except LookupError:

            def on_reassembly_error(error: TransferReassembler.Error) -> None:
                self._statistics.errors += 1
                d = self._statistics.reassembly_errors_per_source_node_id[source_node_id]
                try:
                    d[error] += 1
                except LookupError:
                    d[error] = 1

            self._statistics.reassembly_errors_per_source_node_id.setdefault(source_node_id, {})
            reasm = TransferReassembler(
                source_node_id=source_node_id,
                extent_bytes=self._payload_metadata.extent_bytes,
                on_error_callback=on_reassembly_error,
            )
            self._reassemblers[source_node_id] = reasm
            _logger.debug("%s: New %s (%d total)", self, reasm, len(self._reassemblers))
            return reasm


def _unittest_input_session() -> None:
    from pytest import raises, approx
    from pyuavcan.transport import InputSessionSpecifier, MessageDataSpecifier, Priority, TransferFrom
    from pyuavcan.transport import PayloadMetadata
    from pyuavcan.transport.commons.high_overhead_transport import TransferCRC

    ts = Timestamp.now()
    prio = Priority.SLOW
    dst_nid = 1234

    run_until_complete = asyncio.get_event_loop().run_until_complete
    get_monotonic = asyncio.get_event_loop().time

    nihil_supernum = b"nihil supernum"

    finalized = False

    def do_finalize() -> None:
        nonlocal finalized
        finalized = True

    session_spec = InputSessionSpecifier(MessageDataSpecifier(2345), None)
    payload_meta = PayloadMetadata(100)

    sis = SerialInputSession(
        specifier=session_spec, payload_metadata=payload_meta, loop=asyncio.get_event_loop(), finalizer=do_finalize
    )
    assert sis.specifier == session_spec
    assert sis.payload_metadata == payload_meta
    assert sis.sample_statistics() == SerialInputSessionStatistics()

    assert sis.transfer_id_timeout == approx(SerialInputSession.DEFAULT_TRANSFER_ID_TIMEOUT)
    sis.transfer_id_timeout = 1.0
    with raises(ValueError):
        sis.transfer_id_timeout = 0.0
    assert sis.transfer_id_timeout == approx(1.0)

    assert run_until_complete(sis.receive(get_monotonic() + 0.1)) is None
    assert run_until_complete(sis.receive(0.0)) is None

    def mk_frame(
        transfer_id: int,
        index: int,
        end_of_transfer: bool,
        payload: typing.Union[bytes, memoryview],
        source_node_id: typing.Optional[int],
    ) -> SerialFrame:
        return SerialFrame(
            priority=prio,
            transfer_id=transfer_id,
            index=index,
            end_of_transfer=end_of_transfer,
            payload=memoryview(payload),
            source_node_id=source_node_id,
            destination_node_id=dst_nid,
            data_specifier=session_spec.data_specifier,
        )

    # ANONYMOUS TRANSFERS.
    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=0, end_of_transfer=False, payload=nihil_supernum, source_node_id=None)
    )
    assert sis.sample_statistics() == SerialInputSessionStatistics(
        frames=1,
        errors=1,
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=1, end_of_transfer=True, payload=nihil_supernum, source_node_id=None)
    )
    assert sis.sample_statistics() == SerialInputSessionStatistics(
        frames=2,
        errors=2,
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=nihil_supernum, source_node_id=None)
    )
    assert sis.sample_statistics() == SerialInputSessionStatistics(
        transfers=1,
        frames=3,
        payload_bytes=len(nihil_supernum),
        errors=2,
    )
    assert run_until_complete(sis.receive(0)) == TransferFrom(
        timestamp=ts, priority=prio, transfer_id=0, fragmented_payload=[memoryview(nihil_supernum)], source_node_id=None
    )
    assert run_until_complete(sis.receive(get_monotonic() + 0.1)) is None
    assert run_until_complete(sis.receive(0.0)) is None

    # VALID TRANSFERS. Notice that they are unordered on purpose. The reassembler can deal with that.
    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=1, end_of_transfer=False, payload=nihil_supernum, source_node_id=1111)
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=nihil_supernum, source_node_id=2222)
    )  # COMPLETED FIRST

    assert sis.sample_statistics() == SerialInputSessionStatistics(
        transfers=2,
        frames=5,
        payload_bytes=len(nihil_supernum) * 2,
        errors=2,
        reassembly_errors_per_source_node_id={
            1111: {},
            2222: {},
        },
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts,
        mk_frame(
            transfer_id=0,
            index=3,
            end_of_transfer=True,
            payload=TransferCRC.new(nihil_supernum * 3).value_as_bytes,
            source_node_id=1111,
        ),
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=0, end_of_transfer=False, payload=nihil_supernum, source_node_id=1111)
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts, mk_frame(transfer_id=0, index=2, end_of_transfer=False, payload=nihil_supernum, source_node_id=1111)
    )  # COMPLETED SECOND

    assert sis.sample_statistics() == SerialInputSessionStatistics(
        transfers=3,
        frames=8,
        payload_bytes=len(nihil_supernum) * 5,
        errors=2,
        reassembly_errors_per_source_node_id={
            1111: {},
            2222: {},
        },
    )

    assert run_until_complete(sis.receive(0)) == TransferFrom(
        timestamp=ts, priority=prio, transfer_id=0, fragmented_payload=[memoryview(nihil_supernum)], source_node_id=2222
    )
    assert run_until_complete(sis.receive(0)) == TransferFrom(
        timestamp=ts,
        priority=prio,
        transfer_id=0,
        fragmented_payload=[memoryview(nihil_supernum)] * 3,
        source_node_id=1111,
    )
    assert run_until_complete(sis.receive(get_monotonic() + 0.1)) is None
    assert run_until_complete(sis.receive(0.0)) is None

    # TRANSFERS WITH REASSEMBLY ERRORS.
    sis._process_frame(  # pylint: disable=protected-access
        ts,
        mk_frame(
            transfer_id=1, index=0, end_of_transfer=False, payload=b"", source_node_id=1111  # EMPTY IN MULTIFRAME
        ),
    )

    sis._process_frame(  # pylint: disable=protected-access
        ts,
        mk_frame(
            transfer_id=2, index=0, end_of_transfer=False, payload=b"", source_node_id=1111  # EMPTY IN MULTIFRAME
        ),
    )

    assert sis.sample_statistics() == SerialInputSessionStatistics(
        transfers=3,
        frames=10,
        payload_bytes=len(nihil_supernum) * 5,
        errors=4,
        reassembly_errors_per_source_node_id={
            1111: {
                TransferReassembler.Error.MULTIFRAME_EMPTY_FRAME: 2,
            },
            2222: {},
        },
    )

    assert not finalized
    sis.close()
    assert finalized
    sis.close()  # Idempotency check



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import itertools
import pyuavcan
from ._frame import Frame
from ._common import TransferCRC


FrameType = typing.TypeVar("FrameType", bound=Frame)


def serialize_transfer(
    fragmented_payload: typing.Sequence[memoryview],
    max_frame_payload_bytes: int,
    frame_factory: typing.Callable[[int, bool, memoryview], FrameType],
) -> typing.Iterable[FrameType]:
    r"""
    Constructs an ordered sequence of frames ready for transmission from the provided data fragments.
    Compatible with any high-overhead transport.

    :param fragmented_payload: The transfer payload we're going to be sending.

    :param max_frame_payload_bytes: Max payload per transport-layer frame.

    :param frame_factory: A callable that accepts (frame index, end of transfer, payload) and returns a frame.
        Normally this would be a closure.

    :return: An iterable that yields frames.

    >>> import dataclasses
    >>> from pyuavcan.transport.commons.high_overhead_transport import Frame
    >>> @dataclasses.dataclass(frozen=True)
    ... class MyFrameType(Frame):
    ...     pass    # Transport-specific definition goes here.
    >>> priority = pyuavcan.transport.Priority.NOMINAL
    >>> transfer_id = 12345
    >>> def construct_frame(index: int, end_of_transfer: bool, payload: memoryview) -> MyFrameType:
    ...     return MyFrameType(priority=priority,
    ...                        transfer_id=transfer_id,
    ...                        index=index,
    ...                        end_of_transfer=end_of_transfer,
    ...                        payload=payload)
    >>> frames = list(serialize_transfer(
    ...     fragmented_payload=[
    ...         memoryview(b'He thought about the Horse: '),         # The CRC of this quote is 0xDDD1FF3A
    ...         memoryview(b'how was she doing there, in the fog?'),
    ...     ],
    ...     max_frame_payload_bytes=53,
    ...     frame_factory=construct_frame,
    ... ))
    >>> frames
    [MyFrameType(..., index=0, end_of_transfer=False, ...), MyFrameType(..., index=1, end_of_transfer=True, ...)]
    >>> bytes(frames[0].payload)    # 53 bytes long, as configured.
    b'He thought about the Horse: how was she doing there, '
    >>> bytes(frames[1].payload)    # The stuff at the end is the four bytes of multi-frame transfer CRC.
    b'in the fog?:\xff\xd1\xdd'
    """
    assert max_frame_payload_bytes > 0
    payload_length = sum(map(len, fragmented_payload))
    if payload_length <= max_frame_payload_bytes:  # SINGLE-FRAME TRANSFER
        payload = fragmented_payload[0] if len(fragmented_payload) == 1 else memoryview(b"".join(fragmented_payload))
        assert len(payload) == payload_length
        assert max_frame_payload_bytes >= len(payload)
        yield frame_factory(0, True, payload)
    else:  # MULTI-FRAME TRANSFER
        crc_bytes = TransferCRC.new(*fragmented_payload).value_as_bytes
        refragmented = pyuavcan.transport.commons.refragment(
            itertools.chain(fragmented_payload, (memoryview(crc_bytes),)), max_frame_payload_bytes
        )
        for frame_index, (end_of_transfer, frag) in enumerate(pyuavcan.util.mark_last(refragmented)):
            yield frame_factory(frame_index, end_of_transfer, frag)


def _unittest_serialize_transfer() -> None:
    from pyuavcan.transport import Priority

    priority = Priority.NOMINAL
    transfer_id = 12345678901234567890

    def construct_frame(index: int, end_of_transfer: bool, payload: memoryview) -> Frame:
        return Frame(
            priority=priority, transfer_id=transfer_id, index=index, end_of_transfer=end_of_transfer, payload=payload
        )

    assert [
        construct_frame(0, True, memoryview(b"hello world")),
    ] == list(serialize_transfer([memoryview(b"hello"), memoryview(b" "), memoryview(b"world")], 100, construct_frame))

    assert [
        construct_frame(0, True, memoryview(b"")),
    ] == list(serialize_transfer([], 100, construct_frame))

    hello_world_crc = pyuavcan.transport.commons.crc.CRC32C()
    hello_world_crc.add(b"hello world")

    assert [
        construct_frame(0, False, memoryview(b"hello")),
        construct_frame(1, False, memoryview(b" worl")),
        construct_frame(2, True, memoryview(b"d" + hello_world_crc.value_as_bytes)),
    ] == list(serialize_transfer([memoryview(b"hello"), memoryview(b" "), memoryview(b"world")], 5, construct_frame))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import os
import sys
import time
import gzip
import typing
import pickle
import base64
import pathlib
import logging
import itertools
import dataclasses

import pydsdl
import nunavut
import nunavut.jinja
import nunavut.postprocessors


_AnyPath = typing.Union[str, pathlib.Path]

_TEMPLATE_DIRECTORY: pathlib.Path = pathlib.Path(__file__).absolute().parent / pathlib.Path("_templates")

_OUTPUT_FILE_PERMISSIONS = 0o444
"""
Read-only for all because the files are autogenerated and should not be edited manually.
"""

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class GeneratedPackageInfo:
    path: pathlib.Path
    """
    Path to the directory that contains the top-level ``__init__.py``.
    """

    models: typing.Sequence[pydsdl.CompositeType]
    """
    List of PyDSDL objects describing the source DSDL definitions.
    This can be used for arbitrarily complex introspection and reflection.
    """

    name: str
    """
    The name of the generated package, which is the same as the name of the DSDL root namespace unless
    the name had to be stropped. See ``nunavut.lang.py.PYTHON_RESERVED_IDENTIFIERS``.
    """


def compile(  # pylint: disable=redefined-builtin
    root_namespace_directory: _AnyPath,
    lookup_directories: typing.Optional[typing.List[_AnyPath]] = None,
    output_directory: typing.Optional[_AnyPath] = None,
    allow_unregulated_fixed_port_id: bool = False,
) -> typing.Optional[GeneratedPackageInfo]:
    """
    This function runs the DSDL compiler, converting a specified DSDL root namespace into a Python package.
    In the generated package, nested DSDL namespaces are represented as Python subpackages,
    DSDL types as Python classes, type version numbers as class name suffixes separated via underscores
    (like ``Type_1_0``), constants as class attributes, fields as properties.
    For a more detailed information on how to use generated types, just generate them and read the resulting
    code -- it is made to be human-readable and contains docstrings.

    Generated packages can be freely moved around the file system or even deployed on other systems as long as
    their dependencies are satisfied, which are ``numpy`` and ``pydsdl``.

    Generated packages do not automatically import their nested subpackages. For example, if the application
    needs to use ``uavcan.node.Heartbeat.1.0``, it has to ``import uavcan.node`` explicitly; doing just
    ``import uavcan`` is not sufficient.

    If the source definition contains identifiers, type names, namespace components, or other entities whose
    names are listed in ``nunavut.lang.py.PYTHON_RESERVED_IDENTIFIERS``,
    the compiler applies stropping by suffixing such entities with an underscore ``_``.
    A small subset of applications may require access to a generated entity without knowing in advance whether
    its name is a reserved identifier or not (i.e., whether it's stropped or not). To simplify usage,
    this submodule provides helper functions
    :func:`pyuavcan.dsdl.get_attribute` and :func:`pyuavcan.dsdl.set_attribute` that provide access to generated
    class/object attributes using their original names before stropping.
    Likewise, the function :func:`pyuavcan.dsdl.get_model` can find a generated type even if any of its name
    components are stropped; e.g., a DSDL type ``str.Type.1.0`` would be imported as ``str_.Type_1_0``.
    None of it, however, is relevant for an application that does not require genericity (vast majority of
    applications don't), so a much easier approach in that case is just to look at the generated code and see
    if there are any stropped identifiers in it, and then just use appropriate names statically.

    ..  tip::

        Production applications should compile their DSDL namespaces as part of the package build process.
        This can be done by overriding the ``build_py`` command in ``setup.py`` and invoking this function from there.

    ..  tip::

        Configure your IDE to index the compilation output directory as a source directory to enable code completion.
        For PyCharm: right click the directory --> "Mark Directory as" ->"Sources Root".

    :param root_namespace_directory:
        The source DSDL root namespace directory path. The last component of the path
        is the name of the root namespace. For example, to generate package for the root namespace ``uavcan``,
        the path would be like ``foo/bar/uavcan``.

    :param lookup_directories:
        An iterable of DSDL root namespace directory paths where to search for referred DSDL
        definitions. The format of each path is the same as for the previous parameter; i.e., the last component
        of each path is a DSDL root namespace name. If you are generating code for a vendor-specific DSDL root
        namespace, make sure to provide at least the path to the standard ``uavcan`` namespace directory here.

    :param output_directory:
        The generated Python package directory will be placed into this directory.
        If not specified or None, the current working directory is used.
        For example, if this argument equals ``foo/bar``, and the DSDL root namespace name is ``uavcan``,
        the top-level ``__init__.py`` of the generated package will end up in ``foo/bar/uavcan/__init__.py``.
        The directory tree will be created automatically if it does not exist (like ``mkdir -p``).
        If the destination exists, it will be silently written over.
        Applications that compile DSDL lazily are recommended to shard the output directory by the library
        version number to avoid compatibility issues with code generated by older versions of the library.
        Don't forget to add the output directory to ``PYTHONPATH``.

    :param allow_unregulated_fixed_port_id:
        If True, the compiler will not reject unregulated data types with fixed port-ID.
        If you are not sure what it means, do not use it, and read the UAVCAN specification first.

    :return:
        An instance of :class:`GeneratedPackageInfo` describing the generated package,
        unless the root namespace is empty, in which case it's None.

    :raises:
        :class:`OSError` if required operations on the file system could not be performed;
        :class:`pydsdl.InvalidDefinitionError` if the source DSDL definitions are invalid;
        :class:`pydsdl.InternalError` if there is a bug in the DSDL processing front-end;
        :class:`ValueError` if any of the arguments are otherwise invalid.

    The following table is an excerpt from the UAVCAN specification. Observe that *unregulated fixed port identifiers*
    are prohibited by default, but it can be overridden.

    +-------+---------------------------------------------------+----------------------------------------------+
    |Scope  | Regulated                                         | Unregulated                                  |
    +=======+===================================================+==============================================+
    |Public |Standard and contributed (e.g., vendor-specific)   |Definitions distributed separately from the   |
    |       |definitions. Fixed port identifiers are allowed;   |UAVCAN specification. Fixed port identifiers  |
    |       |they are called *"regulated port-IDs"*.            |are *not allowed*.                            |
    +-------+---------------------------------------------------+----------------------------------------------+
    |Private|Nonexistent category.                              |Definitions that are not available to anyone  |
    |       |                                                   |except their authors. Fixed port identifiers  |
    |       |                                                   |are permitted (although not recommended); they|
    |       |                                                   |are called *"unregulated fixed port-IDs"*.    |
    +-------+---------------------------------------------------+----------------------------------------------+
    """
    started_at = time.monotonic()

    if isinstance(lookup_directories, (str, bytes, pathlib.Path)):
        # https://forum.uavcan.org/t/nestedrootnamespaceerror-in-basic-usage-demo/794
        raise TypeError(f"Lookup directories shall be an iterable of paths, not {type(lookup_directories).__name__}")

    output_directory = pathlib.Path(pathlib.Path.cwd() if output_directory is None else output_directory).resolve()
    root_namespace_directory = pathlib.Path(root_namespace_directory).resolve()
    if root_namespace_directory.parent == output_directory:
        # https://github.com/UAVCAN/pyuavcan/issues/133 and https://github.com/UAVCAN/pyuavcan/issues/127
        raise ValueError(
            "The specified destination may overwrite the DSDL root namespace directory. "
            "Consider specifying a different output directory instead."
        )

    # Read the DSDL definitions
    composite_types = pydsdl.read_namespace(
        root_namespace_directory=str(root_namespace_directory),
        lookup_directories=list(map(str, lookup_directories or [])),
        allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
    )
    if not composite_types:
        _logger.info("Root namespace directory %r does not contain DSDL definitions", root_namespace_directory)
        return None
    (root_namespace_name,) = set(map(lambda x: x.root_namespace, composite_types))  # type: str,
    _logger.info("Read %d definitions from root namespace %r", len(composite_types), root_namespace_name)

    # Template primitives
    filters = {
        "pickle": _pickle_object,
        "numpy_scalar_type": _numpy_scalar_type,
    }

    # Generate code
    assert isinstance(output_directory, pathlib.Path)
    language_context = nunavut.lang.LanguageContext("py", namespace_output_stem="__init__")
    root_ns = nunavut.build_namespace_tree(
        types=composite_types,
        root_namespace_dir=str(root_namespace_directory),
        output_dir=str(output_directory),
        language_context=language_context,
    )
    generator = nunavut.jinja.DSDLCodeGenerator(
        namespace=root_ns,
        generate_namespace_types=nunavut.YesNoDefault.YES,
        templates_dir=_TEMPLATE_DIRECTORY,
        followlinks=True,
        additional_filters=filters,
        post_processors=[
            nunavut.postprocessors.SetFileMode(_OUTPUT_FILE_PERMISSIONS),
            nunavut.postprocessors.LimitEmptyLines(2),
            nunavut.postprocessors.TrimTrailingWhitespace(),
        ],
    )
    generator.generate_all()
    _logger.info(
        "Generated %d types from the root namespace %r in %.1f seconds",
        len(composite_types),
        root_namespace_name,
        time.monotonic() - started_at,
    )

    # A minor UX improvement; see https://github.com/UAVCAN/pyuavcan/issues/115
    for p in sys.path:
        if pathlib.Path(p).resolve() == pathlib.Path(output_directory):
            break
    else:
        if os.name == "nt":
            quick_fix = f'Quick fix: `$env:PYTHONPATH += ";{output_directory.resolve()}"`'
        elif os.name == "posix":
            quick_fix = f'Quick fix: `export PYTHONPATH="{output_directory.resolve()}"`'
        else:
            quick_fix = "Quick fix is not available for this OS."
        _logger.info(
            "Generated package is stored in %r, which is not in Python module search path list. "
            "The package will fail to import unless you add the destination directory to sys.path or PYTHONPATH. %s",
            str(output_directory),
            quick_fix,
        )

    return GeneratedPackageInfo(
        path=pathlib.Path(output_directory) / pathlib.Path(root_namespace_name),
        models=composite_types,
        name=root_namespace_name,
    )


def compile_all(
    root_namespace_directories: typing.Iterable[_AnyPath],
    output_directory: typing.Optional[_AnyPath] = None,
    *,
    allow_unregulated_fixed_port_id: bool = False,
) -> typing.List[GeneratedPackageInfo]:
    """
    This is a simple convenience wrapper over :func:`compile` that addresses a very common use case
    where the application needs to compile multiple inter-dependent namespaces.

    :param root_namespace_directories:
        :func:`compile` will be invoked once for each directory in the list,
        using all of them as look-up dirs for each other.
        They may be ordered arbitrarily.
        Directories that contain no DSDL definitions are ignored.

    :param output_directory:
        See :func:`compile`.

    :param allow_unregulated_fixed_port_id:
        See :func:`compile`.

    :return:
        A list of of :class:`GeneratedPackageInfo`, one per non-empty root namespace directory.

    ..  doctest::
        :hide:

        >>> from tests import DEMO_DIR
        >>> original_sys_path = sys.path
        >>> sys.path = [x for x in sys.path if "compiled" not in x]

    >>> import sys
    >>> import pathlib
    >>> import importlib
    >>> import pyuavcan
    >>> compiled_dsdl_dir = pathlib.Path(".lazy_compiled", pyuavcan.__version__)
    >>> compiled_dsdl_dir.mkdir(parents=True, exist_ok=True)
    >>> sys.path.insert(0, str(compiled_dsdl_dir))
    >>> try:
    ...     import sirius_cyber_corp
    ...     import uavcan.si.sample.volumetric_flow_rate
    ... except (ImportError, AttributeError):
    ...     _ = pyuavcan.dsdl.compile_all(
    ...         [
    ...             DEMO_DIR / "custom_data_types/sirius_cyber_corp",
    ...             DEMO_DIR / "public_regulated_data_types/uavcan",
    ...             DEMO_DIR / "public_regulated_data_types/reg/",
    ...         ],
    ...         output_directory=compiled_dsdl_dir,
    ...     )
    ...     importlib.invalidate_caches()
    ...     import sirius_cyber_corp
    ...     import uavcan.si.sample.volumetric_flow_rate

    ..  doctest::
        :hide:

        >>> sys.path = original_sys_path
    """
    out: typing.List[GeneratedPackageInfo] = []
    root_namespace_directories = list(root_namespace_directories)
    for nsd in root_namespace_directories:
        gpi = compile(
            nsd,
            root_namespace_directories,
            output_directory=output_directory,
            allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
        )
        if gpi is not None:
            out.append(gpi)
    return out


def _pickle_object(x: typing.Any) -> str:
    pck: str = base64.b85encode(gzip.compress(pickle.dumps(x, protocol=4))).decode().strip()
    segment_gen = map("".join, itertools.zip_longest(*([iter(pck)] * 100), fillvalue=""))
    return "\n".join(repr(x) for x in segment_gen)


def _numpy_scalar_type(t: pydsdl.Any) -> str:
    def pick_width(w: int) -> int:
        for o in [8, 16, 32, 64]:
            if w <= o:
                return o
        raise ValueError(f"Invalid bit width: {w}")  # pragma: no cover

    if isinstance(t, pydsdl.BooleanType):
        return "bool"  # numpy.bool is deprecated in v1.20
    if isinstance(t, pydsdl.SignedIntegerType):
        return f"_np_.int{pick_width(t.bit_length)}"
    if isinstance(t, pydsdl.UnsignedIntegerType):
        return f"_np_.uint{pick_width(t.bit_length)}"
    if isinstance(t, pydsdl.FloatType):
        return f"_np_.float{pick_width(t.bit_length)}"
    assert not isinstance(t, pydsdl.PrimitiveType), "Forgot to handle some primitive types"
    return "object"  # numpy.object is deprecated in v1.20



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import pyuavcan.transport


class Deduplicator(abc.ABC):
    """
    The abstract class implementing the transfer-wise deduplication strategy.
    **Users of redundant transports do not need to deduplicate their transfers manually
    as it will be done automatically.**
    Please read the module documentation for further details.
    """

    MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD = int(2 ** 48)
    """
    An inferior transport whose transfer-ID modulo is less than this value is expected to experience
    transfer-ID overflows routinely during its operation. Otherwise, the transfer-ID is not expected to
    overflow for centuries.

    A transfer-ID counter that is expected to overflow is called "cyclic", otherwise it's "monotonic".
    Read https://forum.uavcan.org/t/alternative-transport-protocols/324.
    See :meth:`new`.
    """

    @staticmethod
    def new(transfer_id_modulo: int) -> Deduplicator:
        """
        A helper factory that constructs a :class:`MonotonicDeduplicator` if the argument is not less than
        :attr:`MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD`, otherwise constructs a :class:`CyclicDeduplicator`.
        """
        from . import CyclicDeduplicator, MonotonicDeduplicator

        if transfer_id_modulo >= Deduplicator.MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD:
            return MonotonicDeduplicator()
        return CyclicDeduplicator(transfer_id_modulo)

    @abc.abstractmethod
    def should_accept_transfer(
        self,
        *,
        iface_id: int,
        transfer_id_timeout: float,
        timestamp: pyuavcan.transport.Timestamp,
        source_node_id: typing.Optional[int],
        transfer_id: int,
    ) -> bool:
        """
        The iface-ID is an arbitrary integer that is unique within the redundant group identifying the transport
        instance the transfer was received from.
        It could be the index of the redundant interface (e.g., 0, 1, 2 for a triply-redundant transport),
        or it could be something else like a memory address of a related object.
        Embedded applications usually use indexes, whereas in PyUAVCAN it may be more convenient to use :func:`id`.

        The transfer-ID timeout is specified in seconds. It is used to handle the case of a node restart.
        """
        raise NotImplementedError
